! mcms.f90 -- binary choice model Monte Carlo experiments
!
! Copyright (C) 2009-2015 Jason R. Blevins
! All rights reserved.
!
! Redistribution and use in source and binary forms, with or without
! modification, are permitted provided that the following conditions are met:
! 1. Redistributions of source code must retain the above copyright
!    notice, this list of conditions and the following disclaimer.
! 2. Redistributions in binary form must reproduce the above copyright
!    notice, this list of conditions and the following disclaimer in the
!    documentation  and/or other materials provided with the distribution.
! 3. Neither the names of the copyright holders nor the names of any
!    contributors may be used to endorse or promote products derived from
!    this software without specific prior written permission.
!
! THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
! AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
! IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
! ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
! LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
! CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
! SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
! INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
! CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
! ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
! POSSIBILITY OF SUCH DAMAGE.

program mcms
  use osl_core, wp => osl_wp
  use osl_random
  use osl_stat, only: osl_sample_stats
  use osl_sort, only: osl_quicksort
  use dataset
  use msd_model
  use msb_model
  use utility
  implicit none

  include 'revision.inc'

  !===== User Settings ========================================================!

  !----- Monte Carlo Parameters -----------------------------------------------!

  ! Experiment type
  character(len=2) :: type
  logical :: discrete = .false.

  ! Number of Monte Carlo trials for each sample size, slackness, and level
  integer :: n_mc = 1000

  ! Sample sizes
  integer :: n_samp_sz = 5
  integer :: subsamp_sz_base = 10
  integer, parameter :: samp_sz_base = 250

  ! Slackness sequences
  integer, parameter :: n_seq = 8
  character(len=14) :: seq_name
  real(wp), dimension(:,:), allocatable :: seq_c

  ! Set DO_CR to .FALSE. to bypass calculation of confidence regions
  logical :: do_cr = .true.

  ! Confidence region coverage levels
  integer, parameter :: n_alpha = 4
  real(wp), parameter, dimension(n_alpha) :: alpha &
       = (/ 0.50_wp, 0.25_wp, 0.10_wp, 0.01_wp /)

  ! Subsampling
  integer :: n_ss = 1000

  !----- Model Specification --------------------------------------------------!

  ! Bounds on support of X_1 (continuous model)
  real(wp), dimension(2) :: x1_range

  ! Finite support of X_1 (discrete model)
  real(wp), dimension(:), allocatable :: x1_sup

  ! Finite support of X_2
  real(wp), dimension(:), allocatable :: x2_sup

  ! Population parameters
  real(wp) :: beta0

  ! Simulation over X_1 (continuous model)
  integer, parameter :: nsim = 5000

  !----- Grid Search ----------------------------------------------------------!

  integer :: n_grid = 4001                                 ! Grid size
  real(wp) :: grid_lb = -2.0_wp                            ! Lower grid bounds
  real(wp) :: grid_ub = 2.0_wp                             ! Upper grid bounds
  real(wp) :: grid_step                                    ! Grid stepsize

  !----- Internal Variables ---------------------------------------------------!

  ! Evaluate Q and Q_N on a grid of size N_GRID.
  real(wp), dimension(:), allocatable :: beta_grid         ! Support points
  real(wp), dimension(:), allocatable :: Q_pop             ! Population obj. fn.

  ! Data
  integer :: n                                             ! Sample size
  integer :: m                                             ! Subsample size
  type(dataset_t), dimension(:), allocatable :: dat        ! Datasets
  type(dataset_t), dimension(:,:), allocatable :: dat_ss   ! Subsamples

  ! Population characteristics
  type(msd_model_t) :: mod_d                               ! Model
  type(msb_model_t) :: mod_b                               ! Model
  real(wp), dimension(2) :: id_set                         ! Identified set
  logical, dimension(:), allocatable :: id_set_ind         ! Identified set indices
  character(len=100) :: population                         ! Reporting

  ! Results
  real(wp), dimension(:,:), allocatable :: tau
  real(wp), dimension(:,:,:), allocatable :: est
  real(wp), dimension(:,:,:), allocatable :: cr
  logical, dimension(:,:,:), allocatable :: est_ind
  logical, dimension(:,:,:), allocatable :: cr_ind

  ! Statistics
  real(wp), dimension(:,:,:), allocatable :: est_mean
  real(wp), dimension(:,:), allocatable :: est_cover
  real(wp), dimension(:,:), allocatable :: est_dist
  real(wp), dimension(:,:,:), allocatable :: est_sd
  real(wp), dimension(:,:,:), allocatable :: cr_mean
  real(wp), dimension(:,:), allocatable :: cr_cover
  real(wp), dimension(:,:), allocatable :: cr_dist
  real(wp), dimension(:,:), allocatable :: c_hat

  ! Random number generator
  type(osl_rng_t) :: rng
  type(osl_rng_t) :: rng_ss

  ! Generic indices
  integer :: i, j, k, l, r

  ! Command line arguments
  integer :: ios
  character(len=32) :: arg

  !===== Main Program =========================================================!

  !----- Read Command Line Arguments ------------------------------------------!

  i = 0
  type = ''
  do while (i < command_argument_count())
     i = i + 1
     call get_command_argument(i, arg)

     select case (arg)
     case ('-v', '--version')
        print '(2a)', 'mcms version ', trim(revision)
        stop
     case ('-h', '--help')
        call print_usage()
        stop
     case ('-t', '--type')
        i = i + 1
        call get_command_argument(i, type)
        select case (type)
        case ('c1', 'C1')
           discrete = .false.
           x1_range = (/ 1.0_wp, 2.0_wp /)
           allocate(x2_sup(2))
           x2_sup = (/ -1.0_wp, 1.0_wp /)
           beta0 = -1.0_wp
        case ('c2', 'C2')
           discrete = .false.
           x1_range = (/ -1.0_wp, 1.0_wp /)
           allocate(x2_sup(2))
           x2_sup = (/ -1.0_wp, 1.0_wp /)
           beta0 = -0.3_wp
        case ('d1', 'D1')
           ! Specification D1
           discrete = .true.
           allocate(x1_sup(5), x2_sup(5))
           x1_sup = (/ -2.0_wp, -1.0_wp, 0.0_wp, 1.0_wp, 2.0_wp /)
           x2_sup = (/ -2.0_wp, -1.0_wp, 0.0_wp, 1.0_wp, 2.0_wp /)
           beta0 = -0.3_wp
        case default
           call print_usage()
           stop
        end select
     case ('-n', '--nmc')
        i = i + 1
        call get_command_argument(i, arg)
        read(arg, *, iostat=ios) n_mc
        if (ios /= 0) then
           print '(a,a,/)', 'Invalid value for argument nmc: ', arg
           call print_usage()
           stop
        else if (n_mc < 1) then
           print '(a,a,/)', 'Argument to nmc must be a positive integer: ', arg
           call print_usage()
           stop
        end if
     case ('-s', '--nsize')
        i = i + 1
        call get_command_argument(i, arg)
        read(arg, *, iostat=ios) n_samp_sz
        if (ios /= 0) then
           print '(a,a,/)', 'Invalid value for argument nsize: ', arg
           call print_usage()
           stop
        else if (n_samp_sz < 1) then
           print '(a,a,/)', 'Argument to nsize must be a positive integer: ', arg
           call print_usage()
           stop
        end if
     case ('-g', '--ngrid')
        i = i + 1
        call get_command_argument(i, arg)
        read(arg, *, iostat=ios) n_grid
        if (ios /= 0) then
           print '(a,a,/)', 'Invalid value for argument ngrid: ', arg
           call print_usage()
           stop
        else if (n_grid < 1) then
           print '(a,a,/)', 'Argument to ngrid must be a positive integer: ', arg
           call print_usage()
           stop
        end if
     case ('--nss')
        i = i + 1
        call get_command_argument(i, arg)
        read(arg, *, iostat=ios) n_ss
        if (ios /= 0) then
           print '(a,a,/)', 'Invalid value for argument nss: ', arg
           call print_usage()
           stop
        else if (n_ss < 1) then
           print '(a,a,/)', 'Argument to nss must be a positive integer: ', arg
           call print_usage()
           stop
        end if
     case ('--ssbase')
        i = i + 1
        call get_command_argument(i, arg)
        read(arg, *, iostat=ios) subsamp_sz_base
        if (ios /= 0) then
           print '(a,a,/)', 'Invalid value for argument ssbase: ', arg
           call print_usage()
           stop
        else if (subsamp_sz_base < 1) then
           print '(a,a,/)', 'Argument to ssbase must be a positive integer: ', arg
           call print_usage()
           stop
        end if
     case ('--no-cr')
        do_cr = .false.
     case default
        print '(a,a,/)', 'Unrecognized command-line option: ', arg
        call print_usage()
        stop
     end select
  end do

  if (type .eq. '') then
     print '(a,/)', 'Experiment type is required'
     call print_usage()
     stop
  end if

  !----- Preparations ---------------------------------------------------------!

  ! Allocate memory

  allocate(beta_grid(n_grid), Q_pop(n_grid), id_set_ind(n_grid))
  allocate(dat(n_mc))
  allocate(tau(n_seq, n_mc))
  allocate(est(2, n_seq, n_mc))
  allocate(est_ind(n_grid, n_seq, n_mc))
  allocate(est_mean(2, n_seq, n_samp_sz))
  allocate(est_cover(n_seq, n_samp_sz))
  allocate(est_dist(n_seq, n_samp_sz))
  allocate(est_sd(2, n_seq, n_samp_sz))
  allocate(seq_c(n_seq, n_mc))

  if (do_cr) then
     allocate(cr(2, n_alpha, n_mc))
     allocate(cr_ind(n_grid, n_alpha, n_mc))
     allocate(c_hat(n_alpha, n_mc))
     allocate(cr_mean(2, n_alpha, n_samp_sz))
     allocate(cr_cover(n_alpha, n_samp_sz))
     allocate(cr_dist(n_alpha, n_samp_sz))
  end if

  ! Seed the PRNG
  call osl_rng_seed(rng, 480923013)
  call osl_rng_seed(rng_ss, 248129)

  ! Initialize the model
  if (discrete) then
     call msd_model_init(mod_d, x1_sup, x2_sup, beta0)
  else
     call msb_model_init(mod_b, x1_range, x2_sup, beta0, nsim)
  end if

  !----- Determine \Theta_I ---------------------------------------------------!

  ! Construct a grid for theta.
  beta_grid = osl_linspace(grid_lb, grid_ub, n_grid)
  grid_step = (grid_ub - grid_lb) / real(n_grid, wp)

  ! Obtain identified set by evaluating population objective function on grid.
  Q_pop = eval_population_objective(beta_grid)
  call get_contour_set(beta_grid, Q_pop, maxval(Q_pop) - epsilon(1.0_wp), &
       id_set_ind, id_set)

  !----- Print Summary --------------------------------------------------------!

  call print_summary()

  !----- Monte Carlo Experiments ----------------------------------------------!

  ! For each sample size
  do i = 1, n_samp_sz
     call tic()
     n = samp_sz_base * 2**(3*(i-1))              ! Double sample size thrice
     m = subsamp_sz_base * i                      ! Increase subsample size linearly
     call print_experiment()
     if (do_cr) then
        allocate(dat_ss(n_ss, n_mc))
     end if

     ! Seed the PRNGs for each sample size for reproducibility
     call osl_rng_seed(rng, 480923013 + 42 * i)
     call osl_rng_seed(rng_ss, 248129 + 42 * i)

     !$OMP PARALLEL DO
     do r = 1, n_mc
        ! Generate all data and subsamples in a critical region, using
        ! a common PRNG, so that the results will be reproducible when
        ! running in parallel.  Although the order in which datasets
        ! are generated and processed may differ, summary statistics
        ! will be unchanged.

        !$OMP CRITICAL
        call dataset_init(dat(r), n)
        if (discrete) then
           call msd_model_generate_data(mod_d, rng, dat(r))
        else
           call msb_model_generate_data(mod_b, rng, dat(r))
        end if
        if (do_cr) then
           do j = 1, n_ss
              call dataset_init(dat_ss(j,r), m)
              call subsample_dat(dat(r), m, rng_ss, dat_ss(j,r))
           end do
        end if
        !$OMP END CRITICAL

        ! Perform a single Monte Carlo trial and store the results.
        call run_experiment(r)

        ! Free memory
        call dataset_free(dat(r))
        if (do_cr) then
           do j = 1, n_ss
              call dataset_free(dat_ss(j,r))
           end do
        end if
     end do
     !$OMP END PARALLEL DO

     ! Average estimate, coverage, and mean Hausdorff distance.
     do l = 1, n_seq
        call osl_sample_stats(est(:,l,:), MEAN=est_mean(:,l,i), SD=est_sd(:,l,i))
        est_cover(l,i) = 0.0_wp
        est_dist(l,i) = 0.0_wp
        do r = 1, n_mc
           if (cover_p(est_ind(:,l,r), id_set_ind)) then
              est_cover(l,i) = est_cover(l,i) + 1.0_wp
           end if
           est_dist(l,i) = est_dist(l,i) + hausdorff(est(:,l,r), id_set)
        end do
        est_cover(l,i) = est_cover(l,i) / real(n_mc, wp)
        est_dist(l,i) = est_dist(l,i) / real(n_mc, wp)
     end do

     ! Average confidence region, coverage, and Hausdorff distance.
     if (do_cr) then
        do k = 1, n_alpha
           call osl_sample_stats(cr(:,k,:), MEAN=cr_mean(:,k,i))
           cr_cover(k,i) = 0.0_wp
           cr_dist(k,i) = 0.0_wp
           do r = 1, n_mc
              if (cover_p(cr_ind(:,k,r), id_set_ind)) then
                 cr_cover(k,i) = cr_cover(k,i) + 1.0_wp
              end if
              cr_dist(k,i) = cr_dist(k,i) + hausdorff(cr(:,k,r), id_set)
           end do
           cr_cover(k,i) = cr_cover(k,i) / real(n_mc, wp)
           cr_dist(k,i) = cr_dist(k,i) / real(n_mc, wp)
        end do
     end if

     do l = 1, n_seq
        call get_seq(l, NAME=seq_name)
        print '("%")'
        print '("% tau_n = ", f6.3, " (", a,")")', &
             sum(tau(l,:)) / real(n_mc, wp), &
             seq_name
        print '("% est_cover = ", f6.3)', est_cover(l,i)
        print '("% est_dist = ", f6.3)', est_dist(l,i)
     end do

     if (do_cr) then
        print '("% c_hat = ", 100f6.3)', sum(c_hat, DIM=2) / real(n_mc, wp)
        print '("% cr_cover = ", 100f6.3)', cr_cover(:,i)
        deallocate(dat_ss)
     end if

     ! Stop timer
     call toc()
  end do

  call print_estimates()
  if (do_cr) then
     call print_confidence_regions()
  end if

  !----- Termination ----------------------------------------------------------!

  ! Free memory
  if (discrete) then
     call msd_model_free(mod_d)
  else
     call msb_model_free(mod_b)
  end if
  deallocate(beta_grid, Q_pop, id_set_ind)
  deallocate(dat)
  deallocate(est)
  deallocate(est_ind)
  deallocate(est_mean)
  deallocate(est_cover)
  deallocate(est_dist)
  deallocate(est_sd)
  deallocate(seq_c)
  if (do_cr) then
     deallocate(cr)
     deallocate(cr_ind)
     deallocate(c_hat)
     deallocate(cr_mean)
     deallocate(cr_cover)
     deallocate(cr_dist)
  end if
  if (allocated(x1_sup)) then
     deallocate(x1_sup)
  end if
  if (allocated(x2_sup)) then
     deallocate(x2_sup)
  end if

contains

  !===== Internal Procedures ==================================================!

  subroutine run_experiment(r)
    integer, intent(in) :: r

    ! Local variables
    real(wp), dimension(n_grid) :: Qn, Qm         ! Objective function
    real(wp), dimension(n_grid) :: Qn_sort        ! Objective function
    real(wp), dimension(n_ss) :: Rn               ! Subsampled estimates of Rn
    real(wp) :: sup_Qn                            ! Maximum of the objective functions
    real(wp) :: inf_Qn                            ! Minimum of the objective functions
    real(wp) :: rate_n, rate_m                    ! Minimum of the objective functions
    integer :: n                                  ! Sample size
    integer :: m                                  ! Subsample size
    integer :: j, k, l                            ! Generic indices

    ! Sample and subsample sizes
    n = dat(r)%n
    if (discrete) then
       rate_n = real(n, wp)**(1.0_wp / 2.0_wp)
    else
       rate_n = real(n, wp)**(2.0_wp / 3.0_wp)
    end if
    if (do_cr) then
       m = dat_ss(1,r)%n
       if (discrete) then
          rate_m = real(m, wp)**(1.0_wp / 2.0_wp)
       else
          rate_m = real(m, wp)**(2.0_wp / 3.0_wp)
       end if
    end if

    ! Evaluate the rescaled sample objective function on the grid.
    Qn = eval_sample_objective(dat(r), beta_grid)

    ! Determine the maximum
    sup_Qn = maxval(Qn)

    ! For each slackness sequence (each constant term)
    do l = 1, n_seq
       ! Slackness
       call get_seq(l, r, n, Qn, tau(l,r))

       ! Obtain an initial estimate of the identified set or use truth
       if (tau(l,r) < 0.0_wp) then
          ! Use true identified set
          est(:,l,r) = id_set
          est_ind(:,l,r) = id_set_ind
       else
          ! Use the estimated set
          call get_contour_set(beta_grid, Qn, sup_Qn - tau(l,r), est_ind(:,l,r), est(:,l,r))
       end if
    end do

    ! Determine Rn for each subsample using the entire parameter space.
    if (do_cr) then

       ! Construct confidence regions for several values of alpha.
       do k = 1, n_alpha

          ! Initialize to use entire parameter space for S_1
          cr_ind(:,k,r) = .true.
          cr(:,k,r) = (/ grid_lb, grid_ub /)

          do while (.true.)
             ! Calculate c_hat for old set
             do j = 1, n_ss
                ! Determine the maximum over the entire parameter space
                Qm = eval_sample_objective(dat_ss(j,r), beta_grid)

                ! Calculate the statistic and rescale
                Rn(j) = rate_m * (maxval(Qm) - minval(Qm, MASK=cr_ind(:,k,r)))
             end do

             ! Obtain the 1 - alpha quantile of the Rn values
             call osl_quicksort(Rn)
             c_hat(k,r) = quantile(Rn, 1.0_wp - alpha(k))

             ! Determine inf_{S_cur} Q_n
             inf_Qn = minval(Qn, MASK=cr_ind(:,k,r))

             ! Stopping criteria
             if (rate_n*(sup_Qn-inf_Qn) <= c_hat(k,r)) then
                exit
             else
                ! Obtain the new confidence region for next loop
                call get_contour_set(beta_grid, Qn, sup_Qn - c_hat(k,r) / rate_n, cr_ind(:,k,r), cr(:,k,r))
             end if
          end do

       end do

    end if

  end subroutine run_experiment

  ! Draw a random subsample of size M from the given dataset.
  subroutine subsample_dat(dat, m, rng, dat_ss)
    type(dataset_t), intent(in) :: dat
    integer, intent(in) :: m
    type(osl_rng_t), intent(inout) :: rng
    type(dataset_t), intent(inout) :: dat_ss
    integer, dimension(m) :: idx

    ! obtain random subsample indices
    idx = sample_without_replacement(rng, dat%n, dat_ss%n)

    ! populate the dat_ss dataset (assume it has been initialized)
    dat_ss%y = dat%y(idx)
    dat_ss%x = dat%x(:,idx)
  end subroutine subsample_dat

  ! Evaluate the population objective function over the given grid.
  function eval_population_objective(grid) result(Q)
    real(wp), dimension(n_grid), intent(in) :: grid
    real(wp), dimension(n_grid) :: Q
    integer :: k

    do k = 1, n_grid
       if (discrete) then
          call msd_model_population_objective(mod_d, grid(k), Q(k))
       else
          call msb_model_population_objective(mod_b, grid(k), Q(k))
       end if
    end do
  end function eval_population_objective

  ! Evaluates the sample objective function with the given dataset
  ! over a fixed grid.
  function eval_sample_objective(dat, grid) result(Q)
    type(dataset_t), intent(in) :: dat
    real(wp), dimension(n_grid), intent(in) :: grid
    real(wp), dimension(n_grid) :: Q
    integer :: k

    do k = 1, n_grid
       if (discrete) then
          call msd_model_sample_objective(mod_d, dat, grid(k), Q(k))
       else
          call msb_model_sample_objective(mod_b, dat, grid(k), Q(k))
       end if
    end do
  end function eval_sample_objective

  ! Obtain a random subset of size K from N elements without
  ! replacement.  Based on Algorithm 3.4.2S of Knuth, Seminumeric
  ! Algorithms.
  function sample_without_replacement(rng, n, k) result(samples)
    type(osl_rng_t), intent(inout) :: rng
    integer, intent(in) :: n, k
    integer, dimension(k) :: samples
    integer :: t, m
    real(wp) :: u

    t = 0                            ! total input records processed
    m = 0                            ! number of items selected so far

    do while (m < k)
       u = osl_rng_uniform(rng)
       if ( (n - t)*u >= k - m ) then
          t = t + 1
       else
          m = m + 1
          t = t + 1
          samples(m) = t
       end if
    end do
  end function sample_without_replacement

  ! Prints a summary of parameters on startup.
  subroutine print_summary()
    print '(a)', '% Binary Choice Model Monte Carlo Experiments'
    print '(a)', '% ==========================================='
    print '(a)', '%'
    print '(a)', '% Revision ' // trim(revision) // '.'
    print '(a)', '%'
    print '(a)', '% Population'
    print '(a)', '% ----------'
    print '(a)', '%'

    if (discrete) then
       call print_vec('% x_1 ~ Uniform({ ', x1_sup, 'f4.1', ', ', ' })')
    else
       call print_vec('% x_1 ~ Uniform(', x1_range, 'f4.1', ', ', ')')
    end if
    call print_vec('% x_2 ~ Uniform({ ', x2_sup, 'f4.1', ', ', ' })')
    call print_vec('% \theta_0 = ( ', (/ beta0 /), 'f5.2', ', ', ' )')
    call print_vec('% \theta_I = [ ', id_set, 'f5.2', ', ', ' ]')

    print '(a)',    '%'
    print '(a)',    '% Monte Carlo Experiments'
    print '(a)',    '% -----------------------'
    print '(a)',    '%'
    print '(a,i0)', '% Number of trials: ', n_mc
    print '(a,i0)', '% Number of sample sizes: ', n_samp_sz
    if (do_cr) then
       print '(a,i0)', '% Number of subsamples: ', n_ss
    end if

    call print_vec('% Grid interval: [ ', (/ grid_lb, grid_ub /), 'f5.2', ', ', ']')

    print '(a,i0)',        '% Number of grid points: ', n_grid
    print '(a,f7.4)',      '% Grid resolution: ', grid_step
    if (.not. discrete) then
       print '(a,i0)',     '% Number of simulation draws: ', nsim
    end if

    ! Format population values for later
    write (population, '(a,f6.3,a,a,f6.3,a,f6.3,a)') &
         '$\theta_0 = ', beta0, '$, ', &
         '$\Theta_I = [', id_set(1), ', ', id_set(2), ']$.'

    print '(a)', '%'
    print '(a)', '% Running experiments...'
    print '(a)', '%'
  end subroutine print_summary

  subroutine print_experiment()
    if (do_cr) then
       print '("% n = ",i0,", m = ",i0,", n_ss = ",i0)', n, m, n_ss
    else
       print '("% n = ",i0)', n
    end if
  end subroutine print_experiment

  ! Print table summarizing estimates.
  subroutine print_estimates()
    print '(a)', '%'
    print '(a)', '% Estimates'
    print '(a)', '% ---------'
    print '(a)', ''
    print '(a)', '\begin{table}[tb]'
    print '(a)', '  \centering'
    print '(a)', '  \begin{tabular}{rrccrr}'
    print '(a)', '    \hline'
    print '(a)', '    \hline'
    print '(a)', '    $\tau_n$ & $n$   & Mean $\hat\Theta_n$ & St. Dev. & $d_H$ \\'
    do l = 1, n_seq
       call get_seq(l, NAME=seq_name)
       print '("    \hline")'
       print '("    \multirow{",i0,"}{*}{", a20, "}")', n_seq, seq_name
       do i = 1, n_samp_sz
          write (*, '("    & ",i8," & ")', advance='no') samp_sz_base * 2**(3*(i-1))
          write (*, '("[ ",f6.3,", ",f6.3," ]  & ")', advance='no') est_mean(:,l,i)
          write (*, '("[ ",f6.3,", ",f6.3," ] &  ")', advance='no') est_sd(:,l,i)
          write (*, '(f6.3," \\")') est_dist(l,i)
       end do
    end do
    print '(a)', '    \hline'
    print '(a)', '  \end{tabular}'
    print '(a)', '  \caption{Estimates. ' // trim(population) // ' }'
    if (discrete) then
       print '(a)', '  \label{tab:mcms:est}'
    else
       print '(a)', '  \label{tab:mcms:est}'
    end if
    print '(a)', '\end{table}'
  end subroutine print_estimates

  ! Print table summarizing confidence regions.
  subroutine print_confidence_regions()
    print '(a)', '%'
    print '(a)', '% Confidence Regions'
    print '(a)', '% ------------------'
    print '(a)', ''
    print '(a)', '\begin{table}'
    print '(a)', '  \centering'
    print '(a)', '  \begin{tabular}{lrrrrr}'
    print '(a)', '    \hline'
    print '(a)', '    \hline'
    print '(a)', '    & & \multicolumn{4}{c}{Empirical Coverage} \\'
    print '(a)', '    $\tau_n$ & $n$    &  0.500 &  0.750 &  0.900 &  0.990 \\'
    print '(a)', '    \hline'
    do i = 1, n_samp_sz
       write (*, '("    &   ", i8," & ")', advance='no') samp_sz_base * 2**(3*(i-1))
       write (*, '(f6.3," & ",f6.3," & ",f6.3," & ",f6.3," \\")') cr_cover(:,i)
    end do
    print '(a)', '    \hline'
    print '(a)', '  \end{tabular}'
    print '(a)', '  \caption{Confidence sets}'
    print '(a)', '  \label{tab:mcms:cs}'
    print '(a)', '\end{table}'
    print '(a)', ''
    print '(a)', '%'
    print '(a)', '% Confidence Regions: Miscellaneous'
    print '(a)', '% ---------------------------------'
    print '(a)', '%'
    print '(a)', '% seq              n        0.500              0.750              0.900              0.990'
    do i = 1, n_samp_sz
       write (*, '("% ",i7,2x)', advance='no') samp_sz_base * 2**(3*(i-1))
       write (*, '("[ ",f6.3,",",f6.3," ]  [ ",f6.3,",",f6.3," ]  [ ",f6.3,",",f6.3," ]  [ ",f6.3,",",f6.3," ]")') &
            cr_mean(:,1,i), cr_mean(:,2,i), cr_mean(:,3,i), cr_mean(:,4,i)
    end do
  end subroutine print_confidence_regions

  subroutine print_usage()
     print '(a)', 'Usage: mcms [OPTIONS]'
     print '(a)'
     print '(a)', 'Performs Monte Carlo experiments using the semiparametric'
     print '(a)', 'binary choice model as described in the following paper:'
     print '(a)'
     print '(a)', '    Blevins, Jason R.  Non-Standard Rates of Convergence of'
     print '(a)', '    Criterion-Function-Based Set Estimators for Binary Response'
     print '(a)', '    Models.  Econometrics Journal, forthcoming.'
     print '(a)'
     print '(a)', 'Specification options:'
     print '(a)', '  -t, --type      C1, C2, or D1 (required)'
     print '(a)'
     print '(a)', 'Experiment options:'
     print '(a)', '  -n, --nmc R     number of Monte Carlo replications (default: 1000)'
     print '(a)', '  -s, --nsize S   increase the sample size eightfold S times (default: 5)'
     print '(a)', '  -g, --ngrid G   use G uniform grid points (default: 4001)'
     print '(a)', '      --nss N     number of random subsamples (default: 1000)'
     print '(a)', '      --ssbase N  subsample base size (default: 25)'
     print '(a)', '      --no-cr     do not calculate confidence regions'
     print '(a)'
     print '(a)', 'Other options:'
     print '(a)', '  -h, --help      print this usage information'
     print '(a)', '  -v, --version   print version information'
     print '(a)'
     print '(a)', 'Example:'
     print '(a)'
     print '(a)', '  mcms --nmc 100 --nsize 3 --ssbase 100 --nss 5000 --ngrid 4001 --type C2'
  end subroutine print_usage

  subroutine get_seq(l, r, n, Qn, tau, name)
    integer, intent(in) :: l
    integer, intent(in), optional :: r
    integer, intent(in), optional :: n
    real(wp), dimension(n_grid), intent(in), optional :: Qn
    real(wp), intent(out), optional :: tau
    character(len=*), intent(out), optional :: name

    real(wp), dimension(n_grid) :: Qn_sort
    real(wp) :: sup_Qn, med_Qn
    integer :: ind

    if (present(n) .and. present(Qn) .and. present(tau)) then
       if (l <= 3) then
          ! Determine scaling factors
          Qn_sort = Qn
          call osl_quicksort(Qn_sort)
          sup_Qn = Qn_sort(n_grid)
          ind = maxloc(Qn_sort, DIM=1, MASK=Qn_sort < sup_Qn)
          med_Qn = sup_Qn - Qn_sort(1)

          seq_c(1, r) = 1.0_wp
          seq_c(2, r) = abs(sup_Qn)
          seq_c(3, r) = abs(med_Qn)

          ! Store sequence element
          tau = seq_c(l, r) * real(n, wp)**(-0.49_wp)
       else if (l == 4) then
          tau = 0.0_wp
       else if (l == 5) then
          tau = -1.0_wp
       else if (l == 6) then
          tau = real(n, wp)**(-1.0_wp / 3.0_wp)
       else if (l == 7) then
          tau = 1.0_wp / log(real(n, wp))
       else if (l == 8) then
          tau = 1.0_wp / (2.0_wp * log(log(real(n, wp))))
       end if
    end if

    if (present(name)) then
       select case (l)
       case (1)
          name = '$n^{-0.49}$'
       case (2)
          name = '$S n^{-0.49}$'
       case (3)
          name = '$M n^{-0.49}$'
       case (4)
          name = '$0$'
       case (5)
          name = 'Infeasible'
       case (6)
          name = 'n^{-1/3}$'
       case (7)
          name = '$1/\ln n$'
       case (8)
          name = '$1/(2\ln\ln n)$'
       end select
    end if
  end subroutine get_seq

end program mcms
