! osl_dist.f90 -- probability distributions
!
! Copyright (C) 2005-2010 Jason R. Blevins
! All rights reserved.
!
! This file is part of the Open Scientific Library.
!
! Redistribution and use in source and binary forms, with or without
! modification, are permitted provided that the following conditions are met:
! 1. Redistributions of source code must retain the above copyright
!    notice, this list of conditions and the following disclaimer.
! 2. Redistributions in binary form must reproduce the above copyright
!    notice, this list of conditions and the following disclaimer in the
!    documentation  and/or other materials provided with the distribution.
! 3. Neither the names of the copyright holders nor the names of any
!    contributors may be used to endorse or promote products derived from
!    this software without specific prior written permission.
!
! THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
! AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
! IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
! ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
! LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
! CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
! SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
! INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
! CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
! ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
! POSSIBILITY OF SUCH DAMAGE.

module osl_dist
  use osl_core, wp => osl_wp
  use osl_special
  implicit none

  private :: osl_rbeta

contains

  !--------------------------------------------------------------------------!
  ! Uniform Distribution                                                     !
  !--------------------------------------------------------------------------!

  ! Returns the pdf of the Uniform distribution on [A,B].
  function osl_uniform_pdf(x, a, b, status) result(pdf)
    real(wp), intent(in) :: x, a, b
    real(wp) :: pdf
    integer, intent(out), optional :: status

    if (a > b) then
       call osl_status(OSL_EINVALID, status, 'A > B in OSL_UNIFORM_PDF')
       pdf = 0.0_wp
       return
    end if

    if (x < a .or. x > b) then
       pdf = 0.0_wp
    else
       pdf = 1.0_wp / (b - a)
    end if

    call osl_status(OSL_SUCCESS, status)
  end function osl_uniform_pdf

  ! Returns the cdf of the Uniform distribution on [A,B].
  function osl_uniform_cdf(x, a, b, status) result(cdf)
    real(wp), intent(in) :: x, a, b
    real(wp) :: cdf
    integer, intent(out), optional :: status

    if (a > b) then
       call osl_status(OSL_EINVALID, status, 'A > B in OSL_UNIFORM_CDF')
       cdf = 0.0_wp
       return
    end if

    if (x < a) then
       cdf = 0.0_wp
    else if (x > b) then
       cdf = 1.0_wp
    else
       cdf = x / (b - a)
    end if

    call osl_status(OSL_SUCCESS, status)
  end function osl_uniform_cdf

  ! Returns the inverse cdf, or quantile function, of the Uniform
  ! distribution on [A,B].
  function osl_uniform_inv(p, a, b, status) result(x)
    real(wp), intent(in) :: p, a, b
    real(wp) :: x
    integer, intent(out), optional :: status

    if (a > b) then
       call osl_status(OSL_EINVALID, status, 'A > B in OSL_UNIFORM_INV')
       x = 0.0_wp
       return
    end if

    if (p < 0.0_wp .or. p > 1.0_wp) then
       call osl_status(OSL_EDOM, status, 'invalid quantile in osl_uniform_inv')
       x = 0.0_wp
    else
       x = a + p * (b - a)
       call osl_status(OSL_SUCCESS, status)
    end if
  end function osl_uniform_inv

  !--------------------------------------------------------------------------!
  ! Standard Normal Distribution                                             !
  !--------------------------------------------------------------------------!

  ! Return the value of the probability distribution function of the
  ! standard Normal distribution evaluated at Z.
  function osl_std_normal_pdf(z, status) result(pdf)
    real(wp), intent(in) :: z
    real(wp) :: pdf
    integer, intent(out), optional :: status

    pdf = OSL_SQRT1_2PI * exp((z*z) / (-2.0_wp))
    call osl_status(OSL_SUCCESS, status)
  end function osl_std_normal_pdf

  ! Return the value of the cumulative distribution function of the
  ! standard Normal distribution evaluated at Z.
  function osl_std_normal_cdf(z, status) result(cdf)
    real(wp), intent(in) :: z
    integer, intent(out), optional :: status
    real(wp) :: cdf
    real(wp), parameter :: b1 =  0.31938153_wp
    real(wp), parameter :: b2 = -0.356563782_wp
    real(wp), parameter :: b3 =  1.781477937_wp
    real(wp), parameter :: b4 = -1.821255978_wp
    real(wp), parameter :: b5 =  1.330274429_wp
    real(wp), parameter :: p  =  0.2316419_wp
    real(wp), parameter :: c2 =  0.3989423_wp
    real(wp) :: a, t, b

    if (z > 6.0_wp) then
       cdf = 1.0_wp
    else if (z < -6.0_wp) then
       cdf = 0.0_wp
    else
       a = abs(z)
       t = 1.0_wp / (1.0_wp + a*p)
       b = c2 * exp((-z) * (z / 2.0_wp))
       cdf = 1.0_wp - b * ((((b5*t + b4)*t + b3)*t + b2)*t + b1)*t
       if (z < 0.0_wp) then
          cdf = 1.0_wp - cdf
       end if
    end if
    call osl_status(OSL_SUCCESS, status)
  end function osl_std_normal_cdf

  ! Return the value of the quantile function, or inverse CDF, of the
  ! standard Normal distribution evaluated at P.
  !
  ! Based on the algorithm found here:
  ! http://home.online.no/~pjacklam/notes/invnorm
  function osl_std_normal_inv(p, status) result(z)
    integer, intent(out), optional :: status
    real(wp), intent(in) :: p
    real(wp) :: z

    real(wp), parameter :: p_low = 0.02425_wp
    real(wp), parameter :: p_high = 1.0_wp - p_low

    real(wp), parameter, dimension(6) :: a = &
         (/ -3.969683028665376E+01_wp,  2.209460984245205E+02_wp, &
            -2.759285104469687E+02_wp,  1.383577518672690E+02_wp, &
            -3.066479806614716E+01_wp,  2.506628277459239E+00_wp /)
    real(wp), parameter, dimension(5) :: b = &
         (/ -5.447609879822406E+01_wp,  1.615858368580409E+02_wp, &
            -1.556989798598866E+02_wp,  6.680131188771972E+01_wp, &
            -1.328068155288572E+01_wp /)
    real(wp), parameter, dimension(6) :: c = &
         (/ -7.784894002430293E-03_wp, -3.223964580411365E-01_wp, &
            -2.400758277161838E+00_wp, -2.549732539343734E+00_wp, &
             4.374664141464968E+00_wp,  2.938163982698783E+00_wp /)
    real(wp), parameter, dimension(4) :: d = &
         (/  7.784695709041462E-03_wp,  3.224671290700398E-01_wp, &
             2.445134137142996E+00_wp,  3.754408661907416E+00_wp /)

    real(wp) :: q, r

    if (p < 0.0_wp .or. p > 1.0_wp) then
       z = 0.0_wp
       call osl_status(OSL_EDOM, status, &
            'invalid quantile in osl_std_normal_inv')
       return
    end if

    if (p < p_low) then
       q = sqrt(-2.0_wp * log(p))
       z = (((((c(1)*q+c(2))*q+c(3))*q+c(4))*q+c(5))*q+c(6)) / &
            ((((d(1)*q+d(2))*q+d(3))*q+d(4))*q+1.0_wp)
    else
       if (p <= p_high) then
          q = p - 0.5_wp
          r = q*q
          z = (((((a(1)*r+a(2))*r+a(3))*r+a(4))*r+a(5))*r+a(6))*q / &
               (((((b(1)*r+b(2))*r+b(3))*r+b(4))*r+b(5))*r+1.0_wp)
       else
          if (p < 1.0_wp) then
             q = sqrt(-2.0_wp * log(1.0_wp - p))
             z = -(((((c(1)*q+c(2))*q+c(3))*q+c(4))*q+c(5))*q+c(6)) / &
                  ((((d(1)*q+d(2))*q+d(3))*q+d(4))*q+1.0_wp)
          else
             z = 0.0_wp
          end if
       end if
    end if
    call osl_status(OSL_SUCCESS, status)
  end function osl_std_normal_inv

  !--------------------------------------------------------------------------!
  ! Normal Distribution                                                      !
  !--------------------------------------------------------------------------!

  ! Return the value of the probability distribution function of the
  ! Normal distribution with mean MU and standard deviation SIGMA
  ! evaluated at X.
  function osl_normal_pdf(x, mu, sigma, status) result(pdf)
    real(wp), intent(in) :: x, mu, sigma
    integer, intent(out), optional :: status
    real(wp) :: pdf

    if (sigma < osl_eps_wp) then
       pdf = 0.0_wp
       call osl_status(OSL_EINVALID, status, &
            'SIGMA not positive in OSL_NORMAL_PDF')
       return
    end if

    pdf = osl_std_normal_pdf((x - mu) / sigma) / sigma
    call osl_status(OSL_SUCCESS, status)
  end function osl_normal_pdf

  ! Return the value of the cumulative distribution function of the
  ! Normal distribution with mean MU and standard deviation SIGMA
  ! evaluated at X.
  function osl_normal_cdf(x, mu, sigma, status) result(cdf)
    real(wp), intent(in) :: x, mu, sigma
    integer, intent(out), optional :: status
    real(wp) :: cdf

    if (sigma < osl_eps_wp) then
       cdf = 0.0_wp
       call osl_status(OSL_EINVALID, status, &
            'SIGMA not positive in OSL_NORMAL_CDF')
       return
    end if

    cdf = osl_std_normal_cdf((x - mu) / sigma)
    call osl_status(OSL_SUCCESS, status)
  end function osl_normal_cdf

  ! Return the value of the quantile function, or inverse CDF of the
  ! Normal distribution with mean MU and standard deviation SIGMA
  ! evaluated at P.
  function osl_normal_inv(p, mu, sigma, status) result(x)
    real(wp), intent(in) :: p, mu, sigma
    integer, intent(out), optional :: status
    real(wp) :: x

    if (p < 0.0_wp .or. p > 1.0_wp) then
       x = 0.0_wp
       call osl_status(OSL_EDOM, status, &
            'invalid quantile in OSL_NORMAL_CDF')
       return
    end if

    if (sigma < osl_eps_wp) then
       x = 0.0_wp
       call osl_status(OSL_EINVALID, status, &
            'SIGMA not positive in OSL_NORMAL_INV')
       return
    end if

    x = mu + sigma * osl_std_normal_inv(p)
    call osl_status(OSL_SUCCESS, status)
  end function osl_normal_inv

  !--------------------------------------------------------------------------!
  ! Standard Exponential Distribution                                        !
  !--------------------------------------------------------------------------!

  function osl_std_exponential_pdf(z, status) result(pdf)
    real(wp), intent(in) :: z
    integer, intent(out), optional :: status
    real(wp) :: pdf

    if (z < 0.0_wp) then
       pdf = 0.0_wp
    else
       pdf = exp(-z)
    end if
    call osl_status(OSL_SUCCESS, status)
  end function osl_std_exponential_pdf

  function osl_std_exponential_cdf(z, status) result(cdf)
    real(wp), intent(in) :: z
    integer, intent(out), optional :: status
    real(wp) :: cdf

    if (z < 0.0_wp) then
       cdf = 0.0_wp
    else
       cdf = 1.0_wp - exp(-z)
    end if
    call osl_status(OSL_SUCCESS, status)
  end function osl_std_exponential_cdf

  function osl_std_exponential_inv(p, status) result(x)
    real(wp), intent(in) :: p
    integer, intent(out), optional :: status
    real(wp) :: x

    if (p >= 0.0_wp .and. p < 1.0_wp) then
       x = -log(1.0_wp - p)
       call osl_status(OSL_SUCCESS, status)
    else
       x = 0.0_wp
       call osl_status(OSL_EDOM, status, &
            'invalid quantile in osl_std_exponential_inv')
    end if
  end function osl_std_exponential_inv

  !--------------------------------------------------------------------------!
  ! Exponential Distribution                                                 !
  !--------------------------------------------------------------------------!

  function osl_exponential_pdf(x, lambda, status) result(pdf)
    real(wp), intent(in) :: x
    real(wp), intent(in) :: lambda
    integer, intent(out), optional :: status
    real(wp) :: pdf

    if (lambda < osl_eps_wp) then
       pdf = 0.0_wp
       call osl_status(OSL_EINVALID, status, &
            "rate parameter LAMBDA must be positive in OSL_EXPONENTIAL_PDF")
       return
    end if

    if (x >= 0.0_wp) then
       pdf = lambda * exp(-lambda * x)
    else
       pdf = 0.0_wp
    end if
    call osl_status(OSL_SUCCESS, status)
  end function osl_exponential_pdf

  function osl_exponential_cdf(x, lambda, status) result(cdf)
    real(wp), intent(in) :: x
    real(wp), intent(in) :: lambda
    integer, intent(out), optional :: status
    real(wp) :: cdf

    if (lambda < osl_eps_wp) then
       cdf = 0.0_wp
       call osl_status(OSL_EINVALID, status, &
            "rate parameter LAMBDA must be positive in OSL_EXPONENTIAL_CDF")
       return
    end if

    if (x >= 0.0_wp) then
       cdf = 1.0_wp - exp(-lambda * x)
    else
       cdf = 0.0_wp
    end if
    call osl_status(OSL_SUCCESS, status)
  end function osl_exponential_cdf

  function osl_exponential_inv(p, lambda, status) result(x)
    real(wp), intent(in) :: p
    real(wp), intent(in) :: lambda
    integer, intent(out), optional :: status
    real(wp) :: x

    if (lambda < osl_eps_wp) then
       x = 0.0_wp
       call osl_status(OSL_EINVALID, status, &
            "rate parameter LAMBDA must be positive in OSL_EXPONENTIAL_INV")
       return
    end if

    if (p >= 0.0_wp .and. p < 1.0_wp) then
       x = -1.0_wp * log(1.0_wp - p) / lambda
       call osl_status(OSL_SUCCESS, status)
    else
       x = 0.0_wp
       call osl_status(OSL_EDOM, status, &
            'invalid quantile in osl_exponential_inv')
    end if
  end function osl_exponential_inv

  !--------------------------------------------------------------------------!
  ! Standard Gamma Distribution                                              !
  !--------------------------------------------------------------------------!

  ! Return the pdf of the standard gamma distribution with shape
  ! parameter A evaluated at X.  The scale parameter is fixed at one
  ! for the standard distribution.
  function osl_std_gamma_pdf(x, a, status) result(pdf)
    real(wp), intent(in) :: x
    real(wp), intent(in) :: a
    integer, intent(out), optional :: status
    integer :: stat
    real(wp) :: pdf

    if (a < osl_eps_wp) then
       pdf = 0.0_wp
       call osl_status(OSL_EINVALID, status, &
            "shape parameter must be positive in OSL_STD_GAMMA_PDF")
    else
       if (x < 0.0_wp) then
          pdf = 0.0_wp
       else if (a < 1.0_wp) then
          pdf = x**(a - 1.0_wp) * exp(-x) / osl_gamma(a)
       else
          pdf = exp((a - 1.0_wp) * log(x) - x - osl_log_gamma(a, stat))
          if (stat /= OSL_SUCCESS) then
             call osl_status(OSL_FAILURE, status, &
                  "call to OSL_GAMMA failed in OSL_STD_GAMMA_PDF")
             return
          end if
       end if
       call osl_status(OSL_SUCCESS, status)
    end if
  end function osl_std_gamma_pdf

  ! Computes the cdf of the standard gamma distribution, the integral
  ! of the standard gamma pdf, defined as
  !
  ! $g(x, a) = \frac{1}{\Gamma(a)} \int_x^\infty t^{a-1} e^{-t} dt$
  !
  ! Here, x must be nonnegative and p must be strictly positive.
  ! This function will return zero otherwise.  This implementation is
  ! based on the Fortran 77 implementation of Shea (1988).
  !
  ! Modifications to the original function include using the built-in
  ! C99 error function (erfc) for the normal approximation, using the
  ! built-in C99 log gamma function (lgamma), and removing the ifault
  ! flag.
  !
  ! Reference:
  !
  !     Shea, B. L. (1988).  Algorithm AS 239: Chi-squared and
  !     Incomplete Gamma Integral.  Applied Statistics 37(3), 466-473.
  function osl_std_gamma_cdf(x, a, status) result(cdf)
    real(wp), intent(in) :: x
    real(wp), intent(in) :: a
    integer, intent(out), optional :: status
    integer :: stat
    real(wp) :: cdf

    real(wp), parameter :: elimit = -88.0_wp
    real(wp), parameter :: oflo = 1.0e+37_wp
    real(wp), parameter :: alimit = 1000.0_wp
    real(wp), parameter :: tol = 1.0e-14_wp
    real(wp), parameter :: xbig = 1.0e+08_wp
    real(wp) :: an, arg, b, c, pn1, pn2, pn3, pn4, pn5, pn6, rn, d

    ! Check arguments
    if (a < osl_eps_wp) then
       cdf = 0.0_wp
       call osl_status(OSL_EINVALID, status, &
            "shape parameter must be positive in OSL_STD_GAMMA_CDF")
       return
    end if

    ! Support is nonnegative
    if (x < 0.0_wp) then
       cdf = 0.0_wp
       call osl_status(OSL_SUCCESS, status)
       return
    end if

    ! Base case
    if (x == 0.0_wp) then
       cdf = 0.0_wp
       call osl_status(OSL_SUCCESS, status)
       return
    end if

    ! If a is large, use a normal approximation
    if (alimit < a) then
       pn1 = 3.0_wp * sqrt(a) * &
            ((x / a)**(1.0_wp / 3.0_wp) + 1.0_wp / (9.0_wp * a) - 1.0_wp)
       cdf = osl_std_normal_cdf(pn1, stat)
       if (stat /= OSL_SUCCESS) then
          call osl_status(OSL_FAILURE, status, &
               "call to OSL_STD_NORMAL_CDF failed in OSL_STD_GAMMA_CDF")
       else
          call osl_status(OSL_SUCCESS, status)
       end if
       return
    end if

    ! If X is large set cdf = 1.
    if (xbig < x) then
       cdf = 1.0_wp
       call osl_status(OSL_SUCCESS, status)
       return
    end if

    ! Use Pearson's series expansion.
    if (x <= 1.0_wp .or. x < a) then
        arg = a * log(x) - x - osl_log_gamma(a + 1.0_wp, stat)
        if (stat /= OSL_SUCCESS) then
           cdf = 0.0_wp
           call osl_status(OSL_FAILURE, status, &
                "call to OSL_LOG_GAMMA failed in OSL_STD_GAMMA_CDF")
           return
        end if
        c = 1.0_wp
        cdf = 1.0_wp
        d = a

        do
           d = d + 1.0_wp
           c = c * x / d
           cdf = cdf + c

           if (c <= tol) then
              exit
           end if
        end do

        arg = arg + log(cdf)

        if (elimit <= arg) then
           cdf = exp(arg)
        else
           cdf = 0.0_wp
        end if

    ! Use a continued fraction expansion.
    else
       arg = a * log(x) - x - osl_log_gamma(a, stat)
       if (stat /= OSL_SUCCESS) then
          cdf = 0.0_wp
          call osl_status(OSL_FAILURE, status, &
               "call to OSL_LOG_GAMMA failed in OSL_STD_GAMMA_CDF")
          return
       end if
       d = 1.0_wp - a
       b = d + x + 1.0_wp
       c = 0.0_wp
       pn1 = 1.0_wp
       pn2 = x
       pn3 = x + 1.0_wp
       pn4 = x * b
       cdf = pn3 / pn4

       do
          d = d + 1.0_wp
          b = b + 2.0_wp
          c = c + 1.0_wp
          an = d * c
          pn5 = b * pn3 - an * pn1
          pn6 = b * pn4 - an * pn2

          if (pn6 /= 0.0_wp) then
             rn = pn5 / pn6
             if (abs(cdf - rn) <= min(tol, tol*rn)) then
                exit
             end if
             cdf = rn
          end if
          pn1 = pn3
          pn2 = pn4
          pn3 = pn5
          pn4 = pn6

          ! Re-scale terms in continued fraction if terms are large.
          if (oflo <= abs(pn5)) then
             pn1 = pn1 / oflo
             pn2 = pn2 / oflo
             pn3 = pn3 / oflo
             pn4 = pn4 / oflo
          end if
       end do

       arg = arg + log(cdf)

       if (elimit <= arg) then
          cdf = 1.0_wp - exp(arg)
       else
          cdf = 1.0_wp
       end if
    end if
    call osl_status(OSL_SUCCESS, status)
  end function osl_std_gamma_cdf

  ! Returns the value of x corresponding to the P quantile in the
  ! standard gamma distribution.  This function follows the method
  ! described by Shea (1988), whereby the quantile function of the
  ! chi-square distribution is evaluated at n = 2*a degrees of freedom
  ! and the resulting deviate is divided by 2.
  !
  ! Reference:
  !
  !     Shea, B. L. (1988).  Algorithm AS 239: Chi-squared and
  !     Incomplete Gamma Integral.  Applied Statistics 37(3), 466-473.
  function osl_std_gamma_inv(p, a, status) result(inv)
    real(wp), intent(in) :: p
    real(wp), intent(in) :: a
    integer, intent(out), optional :: status
    integer :: stat
    real(wp) :: inv

    if (p < 0.0_wp .or. p > 1.0_wp) then
       inv = 0.0_wp
       call osl_status(OSL_EDOM, status, &
            "invalid quantile in OSL_STD_GAMMA_INV")
    else if (a < osl_eps_wp) then
       inv = 0.0_wp
       call osl_status(OSL_EINVALID, status, &
            "shape parameter must be positive in OSL_STD_GAMMA_INV")
    else
       inv = 0.5_wp * osl_chisquare_inv(p, 2.0_wp*a, stat)
       if (stat /= OSL_SUCCESS) then
          call osl_status(OSL_FAILURE, status, &
               "call to OSL_CHISQUARE_INV failed in OSL_STD_GAMMA_INV")
       else
          call osl_status(OSL_SUCCESS, status)
       end if
    end if
  end function osl_std_gamma_inv

  !--------------------------------------------------------------------------!
  ! Gamma Distribution                                                       !
  !--------------------------------------------------------------------------!

  ! Evaluates the pdf of the gamma distribution with shape parameter A
  ! and scale parameter B at X.
  function osl_gamma_pdf(x, a, b, status) result(pdf)
    real(wp), intent(in) :: x
    real(wp), intent(in) :: a
    real(wp), intent(in) :: b
    integer, intent(out), optional :: status
    integer :: stat
    real(wp) :: pdf

    if (a < osl_eps_wp) then
       pdf = 0.0_wp
       call osl_status(OSL_EINVALID, status, &
            "shape parameter must be positive in OSL_GAMMA_PDF")
    else if (b < osl_eps_wp) then
       pdf = 0.0_wp
       call osl_status(OSL_EINVALID, status, &
            "scale parameter must be positive in OSL_GAMMA_PDF")
    else
       if (x < 0.0_wp) then
          pdf = 0.0_wp
          call osl_status(OSL_SUCCESS, status)
       else if (a < 1.0_wp) then
          pdf = x**(a - 1.0_wp) * exp(-x / b) / osl_gamma(a) / (b**a)
          call osl_status(OSL_SUCCESS, status)
       else
          pdf = exp((a - 1.0_wp) * log(x / b) - (x / b) &
               - osl_log_gamma(a, stat)) / b
          if (stat /= OSL_SUCCESS) then
             call osl_status(OSL_FAILURE, status, &
                  "call to OSL_LOG_GAMMA failed in OSL_GAMMA_PDF")
          else
             call osl_status(OSL_SUCCESS, status)
          end if
       end if
    end if
  end function osl_gamma_pdf

  ! Evaluates the cdf of the gamma distribution with shape parameter A
  ! and scale parameter B at X.
  function osl_gamma_cdf(x, a, b, status) result(cdf)
    real(wp), intent(in) :: x
    real(wp), intent(in) :: a
    real(wp), intent(in) :: b
    integer, intent(out), optional :: status
    integer :: stat
    real(wp) :: cdf

    if (a < osl_eps_wp) then
       cdf = 0.0_wp
       call osl_status(OSL_EINVALID, status, &
            "shape parameter must be positive in OSL_GAMMA_CDF")
    else if (b < osl_eps_wp) then
       cdf = 0.0_wp
       call osl_status(OSL_EINVALID, status, &
            "scale parameter must be positive in OSL_GAMMA_CDF")
    else
       cdf = osl_std_gamma_cdf(x / b, a, stat)
       if (stat /= OSL_SUCCESS) then
          call osl_status(OSL_FAILURE, status, &
               "call to OSL_STD_GAMMA_CDF failed in OSL_GAMMA_CDF")
       else
          call osl_status(OSL_SUCCESS, status)
       end if
    end if
  end function osl_gamma_cdf

  ! Evaluates the quantile function of the gamma distribution with
  ! shape parameter A and scale parameter B evaluated at P.
  function osl_gamma_inv(p, a, b, status) result(inv)
    real(wp), intent(in) :: p
    real(wp), intent(in) :: a
    real(wp), intent(in) :: b
    integer, intent(out), optional :: status
    integer :: stat
    real(wp) :: inv

    inv = 0.0_wp
    if (p < 0.0_wp .or. p > 1.0_wp) then
       call osl_status(OSL_EDOM, status, &
            "invalid quantile in OSL_GAMMA_INV")
    else if (a < osl_eps_wp) then
       call osl_status(OSL_EINVALID, status, &
            "shape parameter must be positive in OSL_GAMMA_INV")
    else if (b < osl_eps_wp) then
       call osl_status(OSL_EINVALID, status, &
            "scale parameter must be positive in OSL_GAMMA_INV")
    else
       inv = 0.5_wp * b * osl_chisquare_inv(p, 2.0_wp*a, stat)
       if (stat /= OSL_SUCCESS) then
          call osl_status(OSL_FAILURE, status, &
               "call to OSL_CHISQUARE_INV failed in OSL_GAMMA_INV")
       end if
    end if
  end function osl_gamma_inv

  !--------------------------------------------------------------------------!
  ! Chi-Square Distribution                                                  !
  !--------------------------------------------------------------------------!

  function osl_chisquare_pdf(x, v, status) result(pdf)
    real(wp), intent(in) :: x
    real(wp), intent(in) :: v
    integer, intent(out), optional :: status
    integer :: stat
    real(wp) :: pdf

    if (v < osl_eps_wp) then
       pdf = 0.0_wp
       call osl_status(OSL_EINVALID, status, &
            "degrees of freedom must be positive in OSL_CHISQUARE_PDF")
    else
       pdf = osl_gamma_pdf(x, v / 2.0_wp, 2.0_wp, stat)
       if (stat /= OSL_SUCCESS) then
          call osl_status(OSL_FAILURE, status, &
               "call to OSL_GAMMA_PDF failed in OSL_CHISQUARE_PDF")
       else
          call osl_status(OSL_SUCCESS, status)
       end if
    end if
  end function osl_chisquare_pdf

  function osl_chisquare_cdf(x, v, status) result(cdf)
    real(wp), intent(in) :: x
    real(wp), intent(in) :: v
    integer, intent(out), optional :: status
    integer :: stat
    real(wp) :: cdf

    if (v < osl_eps_wp) then
       cdf = 0.0_wp
       call osl_status(OSL_EINVALID, status, &
            "degrees of freedom must be positive in OSL_CHISQUARE_CDF")
    else
       cdf = osl_std_gamma_cdf(x / 2.0_wp, v / 2.0_wp, stat)
       if (stat /= OSL_SUCCESS) then
          call osl_status(OSL_FAILURE, status, &
               "call to OSL_STD_GAMMA_CDF failed in OSL_CHISQUARE_CDF")
       else
          call osl_status(OSL_SUCCESS, status)
       end if
    end if
  end function osl_chisquare_cdf

  ! Evaluates the quantile function of the chi-square distribution.
  ! Note that V, the degrees of freedom, must be positive.
  !
  ! This is an implementation of Applied Statistics Algorithm AS 91
  ! developed by Best and Roberts (1975).  It incorporates the changes
  ! suggested in AS R85.
  !
  ! References:
  !
  !    Best, D. J. and D. E. Roberts (1975).  Algorithm AS 91: The
  !    Percentage Points of the $\chi^2$ Distribution.  Applied
  !    Statistics 24(3), 385-390.
  !
  !    Shea, B. L. (1991).  Algorithm AS R85: A Remark on AS 91: The
  !    Percentage Points of the $\chi^2$ Distribution.  Applied
  !    Statistics 40(1), 233-235.
  function osl_chisquare_inv(p, v, status) result(inv)
    real(wp), intent(in) :: p
    real(wp), intent(in) :: v
    integer, intent(out), optional :: status
    integer :: stat
    real(wp) :: inv

    ! Constants
    integer, parameter ::  maxit = 20
    real(wp), parameter :: aa = 0.6931471806_wp
    real(wp), parameter :: e = 0.5e-06_wp

    real(wp) :: a, b, c, ch, g, p1, p2, s1, s2, s3, s4, s5, s6, t, q, x, xx
    integer :: i

    inv = 0.0_wp
    if (p < 0.0_wp .or. p > 1.0_wp) then
       call osl_status(OSL_EDOM, status, &
            "invalid quantile in OSL_CHISQUARE_INV")
       return
    else if (v < osl_eps_wp) then
       call osl_status(OSL_EINVALID, status, &
            "degrees of freedom must be positive in OSL_CHISQUARE_INV")
       return
    end if
    call osl_status(OSL_SUCCESS, status)

    g = osl_log_gamma(v / 2.0_wp, stat)
    if (stat /= OSL_SUCCESS) then
       call osl_status(OSL_FAILURE, status, &
            "call to OSL_LOG_GAMMA failed in OSL_CHISQUARE_INV")
       return
    end if
    xx = 0.5_wp * v
    c = xx - 1.0_wp

    ! Starting approximation for small chi-squared
    if (v < -1.24_wp * log(p)) then
       ch = (p * xx * exp(g + xx * aa))**(1.0_wp / xx)

       if (ch < e) then
          inv = ch
          return
       end if

    ! Starting approximation for v less than or equal to 0.32
    else if (v <= 0.32_wp) then
       ch = 0.4_wp
       a = log(1.0_wp - p)

       do
          q = ch
          p1 = 1.0_wp + ch * (4.67_wp + ch)
          p2 = ch * (6.73_wp + ch * (6.66_wp * ch))
          t = -0.5_wp + (4.67_wp + 2.0_wp * ch) / p1 &
                   - (6.73_wp + ch * (13.32_wp + 3.0_wp * ch)) / p2
          ch = ch - (1.0_wp - exp(a + g + 0.5_wp * ch + c * aa) * p2 / p1) / t

          if (abs(q / ch - 1.0_wp) <= 0.01_wp) then
             exit
          end if
       end do

    else
       ! Evaluate the inverse normal cdf (e.g., AS 70, AS 111, or AS 241)
       x = osl_std_normal_inv(p, stat)
       if (stat /= OSL_SUCCESS) then
          call osl_status(OSL_FAILURE, status, &
               "call to OSL_STD_NORMAL_INV failed in OSL_CHISQUARE_INV")
          return
       end if

       ! Starting approximation using Wilson and Hilferty estimate
       p1 = 0.222222_wp / v
       ch = v * (x * sqrt(p1) + 1.0_wp - p1)**3

       ! Starting approximation for P tending to 1.
       if (2.2_wp * v + 6.0_wp < ch) then
          ch = -2.0_wp * (log(1.0_wp - p) - c * log(0.5_wp * ch) + g)
       end if
    end if

    ! Call standard gamma cdf (AS 239) and build seven term Taylor series
    do i = 1, maxit
       q = ch
       p1 = 0.5_wp * ch
       p2 = p - osl_std_gamma_cdf(p1, xx, stat)
       if (stat /= OSL_SUCCESS) then
          call osl_status(OSL_FAILURE, status, &
               "call to OSL_STD_GAMMA_CDF failed in OSL_CHISQUARE_INV")
          return
       end if

       t = p2 * exp(xx * aa + g + p1 - c * log(ch))
       b = t / ch
       a = 0.5_wp * t - b * c
       s1 = (210.0_wp+a*(140.0_wp+a*(105.0_wp+a*(84.0_wp+a*(70.0_wp+60.0_wp*a)))))/420.0_wp
       s2 = (420.0_wp+a*(735.0_wp+a*(966.0_wp+a*(1141.0_wp+1278.0_wp*a))))/2520.0_wp
       s3 = (210.0_wp+a*(462.0_wp+a*(707.0_wp+932.0_wp*a)))/2520.0_wp
       s4 = (252.0_wp+a*(672.0_wp+1182.0_wp*a)+c*(294.0_wp+a*(889.0_wp+1740.0_wp*a)))/5040.0_wp
       s5 = (84.0_wp+264.0_wp*a+c*(175.0_wp+606.0_wp*a))/2520.0_wp
       s6 = (120.0_wp+c*(346.0_wp+127.0_wp*c))/5040.0_wp
       ch = ch + t*(1.0_wp+0.5_wp*t*s1-b*c*(s1-b*(s2-b*(s3-b*(s4-b*(s5-b*s6))))))

       if (abs(q / ch - 1.0_wp) > e) then
          inv = ch
          exit
       end if
    end do
  end function osl_chisquare_inv

  !--------------------------------------------------------------------------!
  ! Beta Distribution                                                        !
  !--------------------------------------------------------------------------!

  ! Evaluates the pdf of the beta distribution with shape parameters A
  ! and B at X.
  function osl_beta_pdf(x, a, b, status) result(pdf)
    real(wp), intent(in) :: x
    real(wp), intent(in) :: a
    real(wp), intent(in) :: b
    integer, intent(out), optional :: status
    integer :: stat
    real(wp) :: pdf
    real(wp) :: gab, ga, gb

    if (a < osl_eps_wp .or. b < osl_eps_wp) then
       pdf = 0.0_wp
       call osl_status(OSL_EINVALID, status, &
            "shape parameters must be positive in OSL_BETA_PDF")
       return
    end if

    if (x < 0.0_wp .or. x > 1.0_wp) then
       pdf = 0.0_wp
    else
       gab = osl_log_gamma(a + b, stat)
       if (stat /= OSL_SUCCESS) then
          pdf = 0.0_wp
          call osl_status(OSL_FAILURE, status, &
               "call to OSL_LOG_GAMMA failed in OSL_BETA_PDF")
          return
       end if

       ga = osl_log_gamma(a)
       if (stat /= OSL_SUCCESS) then
          pdf = 0.0_wp
          call osl_status(OSL_FAILURE, status, &
               "call to OSL_LOG_GAMMA failed in OSL_BETA_PDF")
          return
       end if

       gb = osl_log_gamma(b)
       if (stat /= OSL_SUCCESS) then
          pdf = 0.0_wp
          call osl_status(OSL_FAILURE, status, &
               "call to OSL_LOG_GAMMA failed in OSL_BETA_PDF")
          return
       end if

       if (x == 0.0_wp .or. x == 1.0_wp) then
          pdf = exp(gab - ga - gb) * x**(a - 1.0_wp) * (1.0_wp - x)**(b - 1.0_wp)
       else
          pdf = exp(gab - ga - gb + log(x)*(a - 1.0_wp) + log(1.0_wp - x)*(b - 1.0_wp))
       end if
    end if
    call osl_status(OSL_SUCCESS, status)
  end function osl_beta_pdf

  ! Computes the regularized incomplete Beta function evaluated at X
  ! with parameters A and B.  This function is the incomplete beta
  ! function scaled to the [0,1] interval.  On input, BETA must be set
  ! to the logarithm of the complete beta function.
  !
  ! Reference:
  !
  !    Majumder, K. L. and G. P. Bhattacharjee (1973).  Algorithm AS 63:
  !    The incomplete Beta Integral.  Applied Statistics 22(3), 409-411.
  function osl_rbeta(x, a, b, beta, status) result(betain)
    real(wp), intent(in) :: x
    real(wp), intent(in) :: a
    real(wp), intent(in) :: b
    real(wp), intent(in) :: beta
    integer, intent(out), optional :: status
    real(wp) :: betain

    real(wp), parameter :: acu = 0.1e-14_wp
    real(wp) :: ai, cx, aa, asq, bb, rx, temp, term, xx
    logical :: indx
    integer :: ns

    ! Check the input arguments.
    if (a <= 0.0_wp .or. b <= 0.0_wp) then
       betain = 0.0_wp
       call osl_status(OSL_EINVALID, status, 'invalid parameter in osl_rbeta')
       return
    else
       call osl_status(OSL_SUCCESS, status)
    end if

    ! Special cases
    if (x <= 0.0_wp) then
       betain = 0.0_wp
       return
    else if (x >= 1.0_wp) then
       betain = 1.0_wp
       return
    end if

    ! Change tail if necessary and determine S.
    asq = a + b
    cx = 1.0_wp - x

    if (a < asq * x) then
       xx = cx
       cx = x
       aa = b
       bb = a
       indx = .true.
    else
       xx = x
       aa = a
       bb = b
       indx = .false.
    end if

    term = 1.0_wp
    ai = 1.0_wp
    betain = 1.0_wp
    ns = int(bb + cx * asq)

    ! Use Soper's reduction formula.
    rx = xx / cx
    temp = bb - ai
    if (ns == 0) then
       rx = xx
    end if

    do
       term = term * temp * rx / (aa + ai)
       betain = betain + term
       temp = abs(term)

       if (temp <= acu .and. temp <= acu * betain) then
          betain = betain * exp(aa * log(xx) + (bb - 1.0_wp) * log(cx) - beta) / aa
          if (indx) then
             betain = 1.0_wp - betain
          end if
          exit
       end if

       ai = ai + 1.0_wp
       ns = ns - 1

       if (0 <= ns) then
          temp = bb - ai
          if (ns == 0) then
             rx = xx
          end if
       else
          temp = asq
          asq = asq + 1.0_wp
       end if
    end do
  end function osl_rbeta

  ! Calculates the cumulative probability of X for the beta distribution
  ! with shape parameters A and B.  The parameters A and B must be
  ! strictly positive.
  !
  ! References:
  !
  !    Lenth, Russell (1987). Algorithm AS 226: Computing Noncentral
  !    Beta Probabilities. Applied Statistics 36(2), 241-244.
  !
  !    Frick, H. (1990).  Algorithm AS R84: A Remark on Algorithm AS
  !    226: Computing Noncentral Beta Probabilities.  Applied Statistics
  !    39(2), 311-312.
  function osl_beta_cdf(x, a, b, status) result(betanc)
    real(wp), intent(in) :: x
    real(wp), intent(in) :: a
    real(wp), intent(in) :: b
    integer, intent(out), optional :: status
    real(wp) :: betanc

    real(wp), parameter :: lambda = 0.0_wp
    real(wp), parameter :: errmax = 1.0e-7_wp
    real(wp), parameter :: ualpha = 5.0_wp
    integer, parameter :: itrmax = 150

    real(wp) :: ax, beta, c, gx, q, sumq, temp, xj
    integer :: stat

    ! Check parameters
    if (a <= 0.0_wp .or. b <= 0.0_wp) then
       betanc = 0.0_wp
       call osl_status(OSL_EINVALID, status, &
            'shape parameters must be positive in OSL_BETA_CDF')
       return
    end if

    if (x <= 0.0_wp) then
       betanc = 0.0_wp
       call osl_status(OSL_SUCCESS, status)
       return
    else if (x >= 1.0_wp) then
       betanc = 1.0_wp
       call osl_status(OSL_SUCCESS, status)
       return
    end if

    c = 0.5_wp * lambda

    ! Initialize the series.
    beta = osl_log_gamma(a, stat)
    if (stat /= OSL_SUCCESS) then
       betanc = 0.0_wp
       call osl_status(OSL_FAILURE, status, &
            'call to OSL_LOG_GAMMA failed in OSL_BETA_CDF')
       return
    end if

    beta = beta + osl_log_gamma(b, stat)
    if (stat /= OSL_SUCCESS) then
       betanc = 0.0_wp
       call osl_status(OSL_FAILURE, status, &
            'call to OSL_LOG_GAMMA failed in OSL_BETA_CDF')
       return
    end if

    beta = beta - osl_log_gamma(a + b, stat)
    if (stat /= OSL_SUCCESS) then
       betanc = 0.0_wp
       call osl_status(OSL_FAILURE, status, &
            'call to OSL_LOG_GAMMA failed in OSL_BETA_CDF')
       return
    end if

    temp = osl_rbeta(x, a, b, beta, stat)
    if (stat /= OSL_SUCCESS) then
       betanc = 0.0_wp
       call osl_status(OSL_FAILURE, status, &
            'call to OSL_RBETA failed in OSL_BETA_CDF')
       return
    end if

    gx = exp(a * log(x) + b * log(1.0_wp - x) - beta - log(a))

    q = exp(-c)

    xj = 0.0_wp
    ax = q * temp
    sumq = 1.0_wp - q
    betanc = ax

    ! Recur over subsequent terms until convergence is achieved.
    do
       xj = xj + 1.0_wp
       temp = temp - gx
       gx = x * (a + b + xj - 1.0_wp) * gx / (a + xj)
       q = q * c / xj
       sumq = sumq - q
       ax = temp * q
       betanc = betanc + ax

       ! Check for convergence
       if (abs((temp - gx) * sumq) <= errmax) then
          exit
       end if
       if (itrmax < int(xj)) then
          betanc = 0.0_wp
          call osl_status(OSL_FAILURE, status, &
               'iteration failed to converge in OSL_BETA_CDF')
          return
       end if
    end do
    call osl_status(OSL_SUCCESS, status)
  end function osl_beta_cdf

  ! Evaluates the quantile function of the beta distribution with
  ! shape parameters A and B at P.  This implementation uses Newton's
  ! method to find a zero of osl_beta_cdf(x, a, b) - p = 0.
  function osl_beta_inv(p, a, b, status) result(inv)
    real(wp), intent(in) :: p
    real(wp), intent(in) :: a
    real(wp), intent(in) :: b
    integer, intent(out), optional :: status
    real(wp) :: inv, x_old, x_new

    integer, parameter :: MAX_ITER = 10000
    real(wp), parameter :: EPS = osl_eps_wp
    integer :: iter, stat
    real(wp) :: h

    if (p < 0.0_wp .or. p > 1.0_wp) then
       inv = 0.0_wp
       call osl_status(OSL_EDOM, status, &
            "invalid quantile in OSL_BETA_INV")
       return
    else if (a < osl_eps_wp .or. b < osl_eps_wp) then
       inv = 0.0_wp
       call osl_status(OSL_EINVALID, status, &
            "shape parameters must be positive in OSL_BETA_INV")
       return
    end if

    iter = 0
    h = huge(h)
    x_old = a / (a + b)

    do while (abs(h) > EPS .and. iter < MAX_ITER)
       iter = iter + 1

       h = osl_beta_cdf(x_old, a, b, stat) - p
       if (stat /= OSL_SUCCESS) then
          inv = 0.0_wp
          call osl_status(OSL_FAILURE, status, &
               "call to OSL_BETA_CDF failed in OSL_BETA_INV")
          return
       end if

       h = h / osl_beta_pdf(x_old, a, b, stat)
       if (stat /= OSL_SUCCESS) then
          inv = 0.0_wp
          call osl_status(OSL_FAILURE, status, &
               "call to OSL_BETA_PDF failed in OSL_BETA_INV")
          return
       end if

       x_new = x_old - h
       x_new = max(x_new, x_old / 10.0_wp)
       x_new = min(x_new, 1.0_wp - (1.0_wp - x_old) / 10.0_wp)
       x_old = x_new
    end do
    inv = x_new

    if (iter >= MAX_ITER) then
       inv = 0.0_wp
       call osl_status(OSL_EMAXITER, status, &
            "maximum number of iterations reached in OSL_BETA_INV")
    else
       call osl_status(OSL_SUCCESS, status)
    end if
  end function osl_beta_inv

  !--------------------------------------------------------------------------!
  ! Standard Gumbel Distribution (Type I Extreme Value Distribution)         !
  !--------------------------------------------------------------------------!

  ! Returns the pdf of the standard gumbel distribution.
  function osl_std_gumbel_pdf(x, status) result(pdf)
    real(wp), intent(in) :: x
    integer, intent(out), optional :: status
    real(wp) :: pdf, z

    z = exp(-x)
    pdf = z * exp(-z)
    call osl_status(OSL_SUCCESS, status)
  end function osl_std_gumbel_pdf

  ! Returns the cdf of the standard Gumbel distribution.
  function osl_std_gumbel_cdf(x, status) result(cdf)
    real(wp), intent(in) :: x
    integer, intent(out), optional :: status
    real(wp) :: cdf

    cdf = exp(-exp(-x))
    call osl_status(OSL_SUCCESS, status)
  end function osl_std_gumbel_cdf

  ! Returns the inverse cdf of the standard gumbel distribution.
  function osl_std_gumbel_inv(p, status) result(x)
    real(wp), intent(in) :: p
    integer, intent(out), optional :: status
    real(wp) :: x

    if (p < 0.0_wp .or. p > 1.0_wp) then
       x = 0.0_wp
       call osl_status(OSL_EDOM, status, &
            "invalid quantile in OSL_STD_GUMBEL_INV")
    else
       x = -log(-log(p))
       call osl_status(OSL_SUCCESS, status)
    end if
  end function osl_std_gumbel_inv

  !--------------------------------------------------------------------------!
  ! Gumbel Distribution (Type I Extreme Value Distribution)                  !
  !--------------------------------------------------------------------------!

  ! Returns the pdf of the Gumbel distribution.
  function osl_gumbel_pdf(x, mu, beta, status) result(pdf)
    real(wp), intent(in) :: x, mu, beta
    integer, intent(out), optional :: status
    real(wp) :: pdf, z

    if (beta < osl_eps_wp) then
       pdf = 0.0_wp
       call osl_status(OSL_EINVALID, status, &
            "invalid scale parameter in OSL_GUMBEL_CDF")
    else
       z = exp(-(x - mu) / beta)
       pdf = z * exp(-z) / beta
       call osl_status(OSL_SUCCESS, status)
    end if
  end function osl_gumbel_pdf

  ! Returns the cdf of the Gumbel distribution.
  function osl_gumbel_cdf(x, mu, beta, status) result(cdf)
    real(wp), intent(in) :: x, mu, beta
    integer, intent(out), optional :: status
    real(wp) :: cdf

    if (beta < osl_eps_wp) then
       cdf = 0.0_wp
       call osl_status(OSL_EINVALID, status, &
            "invalid scale parameter in OSL_GUMBEL_CDF")
    else
       cdf = exp(-exp(-(x - mu) / beta))
       call osl_status(OSL_SUCCESS, status)
    end if
  end function osl_gumbel_cdf

  ! Returns the inverse cdf of the Gumbel distribution.
  function osl_gumbel_inv(p, mu, beta, status) result(x)
    real(wp), intent(in) :: p, mu, beta
    integer, intent(out), optional :: status
    real(wp) :: x

    if (p < 0.0_wp .or. p > 1.0_wp) then
       x = 0.0_wp
       call osl_status(OSL_EDOM, status, &
            "invalid quantile in OSL_GUMBEL_INV")
    else if (beta < osl_eps_wp) then
       x = 0.0_wp
       call osl_status(OSL_EINVALID, status, &
            "invalid scale parameter in OSL_GUMBEL_INV")
    else
       x = mu - beta * log(-log(p))
       call osl_status(OSL_SUCCESS, status)
    end if
  end function osl_gumbel_inv

end module osl_dist
