! osl_random.f90 -- pseudo-random number generation
!
! Copyright (C) 2005-2010 Jason R. Blevins
! All rights reserved.
!
! This file is part of the Open Scientific Library.
!
! Redistribution and use in source and binary forms, with or without
! modification, are permitted provided that the following conditions are met:
! 1. Redistributions of source code must retain the above copyright
!    notice, this list of conditions and the following disclaimer.
! 2. Redistributions in binary form must reproduce the above copyright
!    notice, this list of conditions and the following disclaimer in the
!    documentation  and/or other materials provided with the distribution.
! 3. Neither the names of the copyright holders nor the names of any
!    contributors may be used to endorse or promote products derived from
!    this software without specific prior written permission.
!
! THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
! AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
! IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
! ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
! LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
! CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
! SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
! INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
! CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
! ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
! POSSIBILITY OF SUCH DAMAGE.

module osl_random
  use osl_core, wp => osl_wp, int32 => osl_int32
  implicit none

  private

  ! Pseudo-random number generator
  public :: osl_rng_t
  public :: osl_rng_min
  public :: osl_rng_max
  public :: osl_rng_seed
  public :: osl_rng_get
  public :: osl_rng_uniform
  public :: osl_rng_uniform_int

  ! Continuous distributions
  public :: osl_uniform_rnd
  public :: osl_std_normal_rnd
  public :: osl_normal_rnd
  public :: osl_std_exponential_rnd
  public :: osl_exponential_rnd
  public :: osl_std_gamma_rnd
  public :: osl_gamma_rnd
  public :: osl_chisquare_rnd
  public :: osl_beta_rnd
  public :: osl_std_gumbel_rnd
  public :: osl_gumbel_rnd

  ! Discrete distributions
  public :: osl_discrete_rnd_init
  public :: osl_discrete_rnd
  public :: osl_discrete_rnd_free

  ! Mersenne Twister parameters
  integer, parameter :: N = 624
  integer, parameter :: M = 397
  integer(int32), parameter :: DEFAULT_SEED = 5489_int32
  integer(int32), parameter :: MATRIX_A = -1727483681       ! 0x9908b0df
  integer(int32), parameter :: UPPER_MASK = -2147483647 - 1 ! 0x80000000
  integer(int32), parameter :: LOWER_MASK = 2147483647      ! 0x7fffffff
  integer(int32), parameter :: TMASKB = -1658038656         ! 0x9d2c5680
  integer(int32), parameter :: TMASKC = -272236544          ! 0xefc60000

  integer(int32), dimension(0:1), parameter :: mag01 = (/ 0_int32, MATRIX_A /)

  type :: osl_rng_t
     private
     integer(int32), dimension(N) :: mt
     integer(int32) :: mti = N + 2
  end type osl_rng_t

  integer(int32), parameter :: osl_rng_min = -2147483647 - 1  ! -(2^31)
  integer(int32), parameter :: osl_rng_max = 2147483647       ! (2^31 - 1)

  ! Discrete distributions
  type, public :: osl_discrete_rnd_t
     private
     integer :: n                                 ! size
     real(wp), dimension(:), pointer :: F         ! cutoffs
     integer, dimension(:), pointer :: L          ! alias values
  end type osl_discrete_rnd_t

  ! Ziggurat method tables (generated by ziggurat.c)
  integer(int32), parameter, dimension(128) :: KN = (/ &
       1991057938,          0, 1611602771, 1826899878, &
       1918584482, 1969227037, 2001281515, 2023368125, &
       2039498179, 2051788381, 2061460127, 2069267110, &
       2075699398, 2081089314, 2085670119, 2089610331, &
       2093034710, 2096037586, 2098691595, 2101053571, &
       2103168620, 2105072996, 2106796166, 2108362327, &
       2109791536, 2111100552, 2112303493, 2113412330, &
       2114437283, 2115387130, 2116269447, 2117090813, &
       2117856962, 2118572919, 2119243101, 2119871411, &
       2120461303, 2121015852, 2121537798, 2122029592, &
       2122493434, 2122931299, 2123344971, 2123736059, &
       2124106020, 2124456175, 2124787725, 2125101763, &
       2125399283, 2125681194, 2125948325, 2126201433, &
       2126441213, 2126668298, 2126883268, 2127086657, &
       2127278949, 2127460589, 2127631985, 2127793506, &
       2127945490, 2128088244, 2128222044, 2128347141, &
       2128463758, 2128572095, 2128672327, 2128764606, &
       2128849065, 2128925811, 2128994934, 2129056501, &
       2129110560, 2129157136, 2129196237, 2129227847, &
       2129251929, 2129268426, 2129277255, 2129278312, &
       2129271467, 2129256561, 2129233410, 2129201800, &
       2129161480, 2129112170, 2129053545, 2128985244, &
       2128906855, 2128817916, 2128717911, 2128606255, &
       2128482298, 2128345305, 2128194452, 2128028813, &
       2127847342, 2127648860, 2127432031, 2127195339, &
       2126937058, 2126655214, 2126347546, 2126011445, &
       2125643893, 2125241376, 2124799783, 2124314271, &
       2123779094, 2123187386, 2122530867, 2121799464, &
       2120980787, 2120059418, 2119015917, 2117825402, &
       2116455471, 2114863093, 2112989789, 2110753906, &
       2108037662, 2104664315, 2100355223, 2094642347, &
       2086670106, 2074676188, 2054300022, 2010539237 /)

  real(wp), parameter, dimension(128) :: WN = (/ &
       1.729040521542798e-09_wp, 1.268092844700276e-10_wp, 1.689751777318455e-10_wp, 1.986268844247905e-10_wp, &
       2.223243179249995e-10_wp, 2.424493612544893e-10_wp, 2.601613190063206e-10_wp, 2.761198871170396e-10_wp, &
       2.907396281771598e-10_wp, 3.042997041437660e-10_wp, 3.169979521395427e-10_wp, 3.289802052711306e-10_wp, &
       3.403573812183406e-10_wp, 3.512160221366471e-10_wp, 3.616250995056517e-10_wp, 3.716405763495978e-10_wp, &
       3.813085643110598e-10_wp, 3.906675680994882e-10_wp, 3.997501186997691e-10_wp, 4.085839861598440e-10_wp, &
       4.171930964016065e-10_wp, 4.255982353459263e-10_wp, 4.338175973925511e-10_wp, 4.418672181252886e-10_wp, &
       4.497613196266582e-10_wp, 4.575125889458829e-10_wp, 4.651324048140010e-10_wp, 4.726310238481176e-10_wp, &
       4.800177347232567e-10_wp, 4.873009867798748e-10_wp, 4.944884980538973e-10_wp, 5.015873466119616e-10_wp, &
       5.086040482424560e-10_wp, 5.155446229195390e-10_wp, 5.224146519706316e-10_wp, 5.292193275006305e-10_wp, &
       5.359634953312890e-10_wp, 5.426516924820619e-10_wp, 5.492881800346021e-10_wp, 5.558769720760773e-10_wp, &
       5.624218612983588e-10_wp, 5.689264417346550e-10_wp, 5.753941290375603e-10_wp, 5.818281786390898e-10_wp, &
       5.882317020812170e-10_wp, 5.946076817624996e-10_wp, 6.009589843108302e-10_wp, 6.072883727627885e-10_wp, &
       6.135985177054135e-10_wp, 6.198920075155922e-10_wp, 6.261713578149429e-10_wp, 6.324390202435402e-10_wp, &
       6.386973906435736e-10_wp, 6.449488167337383e-10_wp, 6.511956053464698e-10_wp, 6.574400292928599e-10_wp, &
       6.636843339139875e-10_wp, 6.699307433723302e-10_wp, 6.761814667327444e-10_wp, 6.824387038791137e-10_wp, &
       6.887046513100733e-10_wp, 6.949815078551667e-10_wp, 7.012714803513155e-10_wp, 7.075767893185560e-10_wp, &
       7.138996746735849e-10_wp, 7.202424015197486e-10_wp, 7.266072660527047e-10_wp, 7.329966016220864e-10_wp, &
       7.394127849911228e-10_wp, 7.458582428383539e-10_wp, 7.523354585483488e-10_wp, 7.588469793417652e-10_wp, &
       7.653954237992263e-10_wp, 7.719834898384400e-10_wp, 7.786139632098381e-10_wp, 7.852897265828997e-10_wp, &
       7.920137693034098e-10_wp, 7.987891979113536e-10_wp, 8.056192475202170e-10_wp, 8.125072941713968e-10_wp, &
       8.194568682925745e-10_wp, 8.264716694066625e-10_wp, 8.335555822587845e-10_wp, 8.407126945532991e-10_wp, &
       8.479473165218372e-10_wp, 8.552640025776094e-10_wp, 8.626675753519363e-10_wp, 8.701631524574424e-10_wp, &
       8.777561763803284e-10_wp, 8.854524479737278e-10_wp, 8.932581641080369e-10_wp, 9.011799601356605e-10_wp, &
       9.092249579511381e-10_wp, 9.174008205786005e-10_wp, 9.257158144040126e-10_wp, 9.341788803988472e-10_wp, &
       9.427997159666314e-10_wp, 9.515888693998883e-10_wp, 9.605578493831253e-10_wp, 9.697192525453944e-10_wp, &
       9.790869127908901e-10_wp, 9.886760770687724e-10_wp, 9.985036134535425e-10_wp, 1.008588258991447e-09_wp, &
       1.018950916862138e-09_wp, 1.029615015200667e-09_wp, 1.040606943699987e-09_wp, 1.051956589272804e-09_wp, &
       1.063697999193087e-09_wp, 1.075870210164582e-09_wp, 1.088518296060728e-09_wp, 1.101694707813504e-09_wp, &
       1.115461009559716e-09_wp, 1.129890161349322e-09_wp, 1.145069570006724e-09_wp, 1.161105242602235e-09_wp, &
       1.178127560945613e-09_wp, 1.196299505385076e-09_wp, 1.215828698329556e-09_wp, 1.236985629080497e-09_wp, &
       1.260132330060852e-09_wp, 1.285769684420515e-09_wp, 1.314620184967718e-09_wp, 1.347783956221085e-09_wp, &
       1.387063531506704e-09_wp, 1.435740319181638e-09_wp, 1.500865903022299e-09_wp, 1.603094793809112e-09_wp /)

  real(wp), parameter, dimension(128) :: FN = (/ &
       1.000000000000000e+00_wp, 9.635996931270862e-01_wp, 9.362826816850596e-01_wp, 9.130436479717402e-01_wp, &
       8.922816507840261e-01_wp, 8.732430489100695e-01_wp, 8.555006078694506e-01_wp, 8.387836052959896e-01_wp, &
       8.229072113814090e-01_wp, 8.077382946829605e-01_wp, 7.931770117713051e-01_wp, 7.791460859296877e-01_wp, &
       7.655841738977045e-01_wp, 7.524415591746114e-01_wp, 7.396772436726473e-01_wp, 7.272569183441848e-01_wp, &
       7.151515074104986e-01_wp, 7.033360990161581e-01_wp, 6.917891434366751e-01_wp, 6.804918409973341e-01_wp, &
       6.694276673488904e-01_wp, 6.585820000500880e-01_wp, 6.479418211102225e-01_wp, 6.374954773350423e-01_wp, &
       6.272324852499273e-01_wp, 6.171433708188809e-01_wp, 6.072195366251203e-01_wp, 5.974531509445167e-01_wp, &
       5.878370544347066e-01_wp, 5.783646811197631e-01_wp, 5.690299910679509e-01_wp, 5.598274127040869e-01_wp, &
       5.507517931146045e-01_wp, 5.417983550254255e-01_wp, 5.329626593838361e-01_wp, 5.242405726729841e-01_wp, &
       5.156282382440018e-01_wp, 5.071220510755690e-01_wp, 4.987186354709795e-01_wp, 4.904148252838441e-01_wp, &
       4.822076463294852e-01_wp, 4.740943006930169e-01_wp, 4.660721526894561e-01_wp, 4.581387162678721e-01_wp, &
       4.502916436820392e-01_wp, 4.425287152754684e-01_wp, 4.348478302499909e-01_wp, 4.272469983049961e-01_wp, &
       4.197243320495744e-01_wp, 4.122780401026610e-01_wp, 4.049064208072229e-01_wp, 3.976078564938733e-01_wp, &
       3.903808082373146e-01_wp, 3.832238110559012e-01_wp, 3.761354695105626e-01_wp, 3.691144536644722e-01_wp, &
       3.621594953693176e-01_wp, 3.552693848479171e-01_wp, 3.484429675463266e-01_wp, 3.416791412315504e-01_wp, &
       3.349768533135892e-01_wp, 3.283350983728503e-01_wp, 3.217529158759849e-01_wp, 3.152293880650109e-01_wp, &
       3.087636380061811e-01_wp, 3.023548277864835e-01_wp, 2.960021568469330e-01_wp, 2.897048604429598e-01_wp, &
       2.834622082232330e-01_wp, 2.772735029191881e-01_wp, 2.711380791383846e-01_wp, 2.650553022555892e-01_wp, &
       2.590245673962048e-01_wp, 2.530452985073258e-01_wp, 2.471169475123214e-01_wp, 2.412389935454398e-01_wp, &
       2.354109422634791e-01_wp, 2.296323252321161e-01_wp, 2.239026993850084e-01_wp, 2.182216465543054e-01_wp, &
       2.125887730717303e-01_wp, 2.070037094399265e-01_wp, 2.014661100743137e-01_wp, 1.959756531162777e-01_wp, &
       1.905320403191371e-01_wp, 1.851349970089922e-01_wp, 1.797842721232955e-01_wp, 1.744796383307895e-01_wp, &
       1.692208922373650e-01_wp, 1.640078546834204e-01_wp, 1.588403711394793e-01_wp, 1.537183122081817e-01_wp, &
       1.486415742423423e-01_wp, 1.436100800906278e-01_wp, 1.386237799845946e-01_wp, 1.336826525834394e-01_wp, &
       1.287867061959432e-01_wp, 1.239359802028678e-01_wp, 1.191305467076508e-01_wp, 1.143705124488660e-01_wp, &
       1.096560210148403e-01_wp, 1.049872554094213e-01_wp, 1.003644410286559e-01_wp, 9.578784912173144e-02_wp, &
       9.125780082683026e-02_wp, 8.677467189478018e-02_wp, 8.233889824223566e-02_wp, 7.795098251397339e-02_wp, &
       7.361150188411340e-02_wp, 6.932111739357791e-02_wp, 6.508058521306807e-02_wp, 6.089077034804041e-02_wp, &
       5.675266348104985e-02_wp, 5.266740190305101e-02_wp, 4.863629585986780e-02_wp, 4.466086220049142e-02_wp, &
       4.074286807444417e-02_wp, 3.688438878665620e-02_wp, 3.308788614622575e-02_wp, 2.935631744000685e-02_wp, &
       2.569329193593427e-02_wp, 2.210330461592710e-02_wp, 1.859210273701129e-02_wp, 1.516729801054657e-02_wp, &
       1.183947865788486e-02_wp, 8.624484412859885e-03_wp, 5.548995220771345e-03_wp, 2.669629083880923e-03_wp /)

contains

  !--------------------------------------------------------------------------!
  ! Mersenne Twister                                                         !
  !--------------------------------------------------------------------------!

  subroutine osl_rng_seed(self, s)
    type(osl_rng_t), intent(inout) :: self
    integer(int32), intent(in) :: s
    integer :: mti

    self%mt(1) = s
    do mti = 2, N
       self%mt(mti) = 1812433253 &
            * ieor(self%mt(mti - 1), ishft(self%mt(mti - 1), -30)) &
            + mti - 1
    end do
    self%mti = N + 1
  end subroutine osl_rng_seed

  function osl_rng_get(self) result(y)
    type(osl_rng_t), intent(inout) :: self
    integer(int32) :: y
    integer :: kk

    ! generate N words at once
    if (self%mti >= N + 1) then
        if (self%mti == N + 2) then
           ! osl_rng_seed() has not been called, use default seed
           call osl_rng_seed(self, DEFAULT_SEED)
        end if
        do kk = 1, N - M
           y = ior(iand(self%mt(kk), UPPER_MASK), &
                   iand(self%mt(kk+1), LOWER_MASK))
           self%mt(kk) = ieor(ieor(self%mt(kk + M), ishft(y, -1)), &
                              mag01(iand(y, 1_int32)))
        end do
        do kk = N - M + 1, N - 1
           y = ior(iand(self%mt(kk), UPPER_MASK), &
                   iand(self%mt(kk+1), LOWER_MASK))
           self%mt(kk) = ieor(ieor(self%mt(kk + (M - N)), ishft(y, -1)), &
                              mag01(iand(y, 1_int32)))
        end do
        y = ior(iand(self%mt(N), UPPER_MASK), iand(self%mt(1), LOWER_MASK))
        self%mt(N) = ieor(ieor(self%mt(M), ishft(y, -1)), mag01(iand(y, 1_int32)))
        self%mti = 1
    end if

    y = self%mt(self%mti)

    ! Tempering
    y = ieor(y, ishft(y, -11))
    y = ieor(y, iand(ishft(y, 7), TMASKB))
    y = ieor(y, iand(ishft(y, 15), TMASKC))
    y = ieor(y, ishft(y, -18))

    self%mti = self%mti + 1
  end function osl_rng_get

  !--------------------------------------------------------------------------!
  ! Generic RNG Interface                                                    !
  !--------------------------------------------------------------------------!

  ! Returns a uniformly distributed real number on [0,1).
  function osl_rng_uniform(self) result(u)
    type(osl_rng_t), intent(inout) :: self
    real(wp) :: u
    ! 1 / 2^32 = 2.3283064365e-10
    u = 0.5_wp + real(osl_rng_get(self), wp) * 2.3283064365e-10_wp
  end function osl_rng_uniform

  ! Returns a uniformly distributed real number on (0,1).
  function osl_rng_uniform_pos(self) result(u)
    type(osl_rng_t), intent(inout) :: self
    real(wp) :: u
    ! 1 / 2^32 = 2.3283064365e-10
    u = 0.5_wp + (real(osl_rng_get(self), wp) + 0.5_wp) * 2.3283064365e-10_wp
  end function osl_rng_uniform_pos

  ! Return a random integer between A and B (inclusive).  Internally,
  ! this function draws a uniform integer from the underlying
  ! generator, from the full range of the generator, and scales it,
  ! being careful to avoid integer overflow in the range.
  function osl_rng_uniform_int(rng, a, b) result(u)
    type(osl_rng_t), intent(inout) :: rng
    integer, intent(in) :: a, b
    integer :: u
    real(wp) :: scale

    scale = real(osl_rng_max, wp) - real(osl_rng_min, wp)
    scale = (real(b, wp) - real(a, wp) + 1.0_wp) / scale
    do
       u = int(real(a, wp) + scale * real(osl_rng_get(rng), wp) &
            - scale * real(osl_rng_min, wp))
       if (u >= a .and. u <= b) then
          exit
       end if
    end do
  end function osl_rng_uniform_int

  !--------------------------------------------------------------------------!
  ! Uniform Distribution                                                     !
  !--------------------------------------------------------------------------!

  ! Draw a uniform real number on [a,b].  For generating standard uniform
  ! draws on [0,1), OSL_RNG_UNIFORM is more efficient.
  function osl_uniform_rnd(rng, a, b) result(rnd)
    type(osl_rng_t), intent(inout) :: rng
    real(wp), intent(in) :: a, b
    real(wp) :: rnd

    rnd = a + (b - a) * osl_rng_uniform(rng)
  end function osl_uniform_rnd

  !--------------------------------------------------------------------------!
  ! Standard Normal Distribution                                             !
  !--------------------------------------------------------------------------!

  ! Draw from the standard Normal distribution using the Ziggurat method.
  !
  ! Reference:
  !
  !     Marsaglia, George and Wai Wan Tsang (2000).  The Ziggurat
  !     Method for Generating Random Variables. Journal of Statistical
  !     Software 5(8).
  function osl_std_normal_rnd(rng) result(rnd)
    type(osl_rng_t), intent(inout) :: rng
    real(wp) :: rnd
    integer(int32) :: hz, iz

    hz = osl_rng_get(rng)
    iz = iand(hz, 127) + 1

    if (abs(hz) < KN(iz)) then
       rnd = real(hz, wp) * WN(iz)
    else
       rnd = nfix()
    end if

  contains

    real(wp) function nfix()
      real(wp), parameter :: R = 3.442620_wp       ! Start of right tail
      real(wp), parameter :: R_INV = 0.2904764_wp  ! 0.2904764 = 1/R
      real(wp) :: x, y

      do
         x = real(hz, wp) * WN(iz)

         ! iz == 1, handle the base strip
         if (iz == 1) then
            do
               x = -log(osl_rng_uniform(rng)) * R_INV
               y = -log(osl_rng_uniform(rng))
               if (y + y > x * x) then
                  exit
               end if
            end do

            nfix = R + x
            if (hz < 0) then
               nfix = -nfix
            end if
            return
         end if

         ! iz > 0, handle the wedges of other strips
         if (FN(iz) + osl_rng_uniform(rng)*(FN(iz-1) - FN(iz)) < exp(-0.5_wp*x*x)) then
            nfix = x
            return
         end if

         ! initiate, try to exit loop
         hz = osl_rng_get(rng)
         iz = iand(hz, 127) + 1
         if (abs(hz) < KN(iz)) then
            nfix = real(hz, wp) * WN(iz)
            return
         end if
       end do
    end function nfix

  end function osl_std_normal_rnd

  !--------------------------------------------------------------------------!
  ! Normal Distribution                                                      !
  !--------------------------------------------------------------------------!

  ! Draw from the standard Normal distribution using the Marsaglia
  ! Polar method.
  function osl_normal_rnd(rng, mu, sigma) result(rnd)
    type(osl_rng_t), intent(inout) :: rng
    real(wp), intent(in) :: mu, sigma
    real(wp) :: rnd

    rnd = mu + sigma * osl_std_normal_rnd(rng)
  end function osl_normal_rnd

  !--------------------------------------------------------------------------!
  ! Standard Exponential Distribution                                        !
  !--------------------------------------------------------------------------!

  ! Draw from the standard exponential distribution using the inverse
  ! cdf transformation.
  function osl_std_exponential_rnd(rng) result(rnd)
    type(osl_rng_t), intent(inout) :: rng
    real(wp) :: rnd

    rnd = -log(osl_rng_uniform(rng))
  end function osl_std_exponential_rnd

  !--------------------------------------------------------------------------!
  ! Exponential Distribution                                                 !
  !--------------------------------------------------------------------------!

  ! Draw from the exponential distribution with rate parameter LAMBDA
  ! using the inverse cdf transformation.  Note that LAMBDA must be
  ! strictly positive, but this condition is not checked for efficiency.
  function osl_exponential_rnd(rng, lambda) result(rnd)
    type(osl_rng_t), intent(inout) :: rng
    real(wp), intent(in) :: lambda
    real(wp) :: rnd

    rnd = -log(osl_rng_uniform(rng)) / lambda
  end function osl_exponential_rnd

  !--------------------------------------------------------------------------!
  ! Gamma Distribution                                                       !
  !--------------------------------------------------------------------------!

  ! Generate standard Gamma distributed pseudo-random numbers with
  ! shape parameter a using the method of Marsaglia and Tsang (2000).
  !
  ! Reference:
  !
  !     Marsaglia, George and Wai Wan Tsang (2000).  A simple method
  !     for generating gamma variables.  ACM Trans. on Math. Software
  !     26(3), 363-372.
  function osl_std_gamma_rnd(rng, a) result(rnd)
    type(osl_rng_t), intent(inout) :: rng
    real(wp), intent(in) :: a
    real(wp) :: rnd

    real(wp) :: d, c, x, v, u
    real(wp) :: shape

    shape = a

    ! This algorithm generates deviates for a >= 1. Following the
    ! note on Page 371 of Marsaglia and Tsang (2000), for a < 1.0,
    ! generate a gamma(a+1) deviate and transform it accordingly.
    if (a < 1.0_wp) then
       shape = shape + 1.0_wp
    end if

    d = shape - 1.0_wp / 3.0_wp
    c = 1.0_wp / sqrt(9.0_wp * d)

    do
       ! Generate v = (1+cx)^3 where x is random normal; repeat if v <= 0.
       do
          x = osl_std_normal_rnd(rng)
          v = 1.0_wp + c * x
          if (v > 0.0_wp) then
             exit
          end if
       end do
       v = v**3

       ! Generate uniform variable
       u = osl_rng_uniform(rng)
       if (u < 1.0_wp - 0.0331_wp * x**4) then
          rnd = d*v
          exit
       end if
       if (log(u) < 0.5_wp * x**2 + d * (1.0_wp - v + log(v))) then
          rnd = d*v
          exit
       end if
    end do

    ! Transformation of deviates with a < 1.0.
    if (a < 1.0_wp) then
       u = osl_rng_uniform(rng)
       rnd = rnd * u**(1.0_wp / a)
    end if
  end function osl_std_gamma_rnd

  !--------------------------------------------------------------------------!
  ! Gamma Distribution                                                       !
  !--------------------------------------------------------------------------!

  function osl_gamma_rnd(rng, a, b) result(rnd)
    type(osl_rng_t), intent(inout) :: rng
    real(wp), intent(in) :: a
    real(wp), intent(in) :: b
    real(wp) :: rnd

    rnd = b * osl_std_gamma_rnd(rng, a)
  end function osl_gamma_rnd

  !--------------------------------------------------------------------------!
  ! Chi-Square Distribution                                                  !
  !--------------------------------------------------------------------------!

  function osl_chisquare_rnd(rng, v) result(rnd)
    type(osl_rng_t), intent(inout) :: rng
    real(wp), intent(in) :: v
    real(wp) :: rnd

    rnd = 2.0_wp * osl_std_gamma_rnd(rng, v / 2.0_wp)
  end function osl_chisquare_rnd

  !--------------------------------------------------------------------------!
  ! Beta Distribution                                                        !
  !--------------------------------------------------------------------------!

  ! Generate a draw from the beta distribution with shape parameters A
  ! and B.  For small values of a and b (when both are less than one)
  ! Johnk's algorithm is used.  Johnk's algorithm becomes less
  ! efficient for larger values of A and B (Robert and Casella, 2004),
  ! and in this case a relationship with the Gamma distribution is
  ! used instead.
  !
  ! References:
  !
  !     Devroye, Luc (1986). Non-Uniform Random Variate Generation.
  !     Springer-Verlag, New York.
  !
  !     Robert, Christian P. and George Casella (2004). Monte Carlo
  !     Statistical Methods. Springer.
  function osl_beta_rnd(rng, a, b) result(rnd)
    type(osl_rng_t), intent(inout) :: rng
    real(wp), intent(in) :: a
    real(wp), intent(in) :: b
    real(wp) :: rnd
    real(wp) :: u, v, x, y

    if (a <= 1.0_wp .and. b <= 1.0_wp) then
       ! Johnk's algorithm (Devroye, 1986, p. 418)
       do
          u = osl_rng_uniform(rng)
          v = osl_rng_uniform(rng)
          x = u**(1.0_wp / a)
          y = v**(1.0_wp / b)

          if ((x + y) <= 1.0_wp) then
             rnd = x / (x + y)
             exit
          end if
       end do
    else
       ! See Theorem 4.1 of Devroye (1986)
       u = osl_std_gamma_rnd(rng, a)
       v = osl_std_gamma_rnd(rng, b)
       rnd = u / (u + v)
    end if

  end function osl_beta_rnd

  !--------------------------------------------------------------------------!
  ! Gumbel Distribution                                                      !
  !--------------------------------------------------------------------------!

  ! Return a draw from the standard Gumbel, or Type I Extreme Value,
  ! distribution with parameters MU = 0.0 and BETA = 1.0.  This
  ! function generates draws using the inverse CDF transformation.
  function osl_std_gumbel_rnd(self) result(rnd)
    type(osl_rng_t), intent(inout) :: self
    real(wp) :: rnd

    rnd = -log(-log(osl_rng_uniform(self)))
  end function osl_std_gumbel_rnd

  ! Return a draw from the Gumbel, or Type I Extreme Value,
  ! distribution with parameters MU and BETA using the inverse CDF
  ! transformation.
  function osl_gumbel_rnd(self, mu, beta) result(rnd)
    type(osl_rng_t), intent(inout) :: self
    real(wp), intent(in) :: mu
    real(wp), intent(in) :: beta
    real(wp) :: rnd

    rnd = mu + beta * osl_std_gumbel_rnd(self)
  end function osl_gumbel_rnd

  !--------------------------------------------------------------------------!
  ! General Discrete Distributions                                           !
  !--------------------------------------------------------------------------!

  ! Initialize a discrete random number distribution with masses P for
  ! later sampling.
  subroutine osl_discrete_rnd_init(self, p)
    type(osl_discrete_rnd_t), intent(inout) :: self
    real(wp), dimension(:), intent(in) :: p
    real(wp), dimension(size(p)) :: p_
    real(wp) :: psum
    integer :: j, k, ng, ns
    integer, dimension(size(p)) :: G, S

    ! Normalize the weights
    self%n = size(p)
    psum = sum(p)
    if (psum < epsilon(psum)) then
       p_ = 1.0_wp / real(self%n, wp)
    else
       p_ = p / psum
    end if

    ! Allocate memory in the osl_discrete_rnd_t type
    allocate(self%F(self%n))
    allocate(self%L(self%n))

    ! Generate the cutoff and alias value tables
    self%F = real(self%n, wp) * p_

    ! Define the sets G and S
    ng = 0
    ns = 0
    do k = 1, self%n
       if (self%F(k) < 1.0_wp) then
          ns = ns + 1
          S(ns) = k
       else
          ng = ng + 1
          G(ng) = k
       end if
    end do

    do while ((ns > 0) .and. (ng > 0))
       k = G(ng)
       j = S(ns)
       self%L(j) = k
       self%F(k) = self%F(k) - (1.0_wp - self%F(j))
       ! Adjust the sets G and S
       ns = ns - 1                                ! Remove j from S
       if (self%F(k) < 1.0_wp) then
          ng = ng - 1                             ! Remove k from G
          ns = ns + 1                             ! Add k to S
          S(ns) = k
       end if
    end do
  end subroutine osl_discrete_rnd_init

  ! Draw a sample X from a general discrete probability distribution
  ! on { 1, 2, ..., N } with weights given by P.  The weights need not
  ! sum to one because they will be normalized internally.  This
  ! procedure uses the alias method of Walker (1977).
  function osl_discrete_rnd(self, rng) result(rnd)
    type(osl_discrete_rnd_t), intent(inout) :: self
    type(osl_rng_t), intent(inout) :: rng
    integer :: rnd
    real(wp) :: u
    integer :: i

    ! Obtain a Uniform(0,1) draw
    u = osl_rng_uniform(rng)

    ! Obtain a draw from the discrete distribution
    u = real(self%n, wp) * u
    i = ceiling(u)
    if (real(i, wp) - u < self%F(i)) then
       rnd = i
    else
       rnd = self%L(i)
    end if
  end function osl_discrete_rnd

  ! Free memory associated with the given discrete distribution object.
  subroutine osl_discrete_rnd_free(self)
    type(osl_discrete_rnd_t), intent(inout) :: self
    deallocate(self%L)
    deallocate(self%F)
  end subroutine osl_discrete_rnd_free

end module osl_random
