! plotqmsd.f90 -- plots the maximum score objective function
!
! Copyright (C) 2009-2011 Jason R. Blevins
! All rights reserved.
!
! Redistribution and use in source and binary forms, with or without
! modification, are permitted provided that the following conditions are met:
! 1. Redistributions of source code must retain the above copyright
!    notice, this list of conditions and the following disclaimer.
! 2. Redistributions in binary form must reproduce the above copyright
!    notice, this list of conditions and the following disclaimer in the
!    documentation  and/or other materials provided with the distribution.
! 3. Neither the names of the copyright holders nor the names of any
!    contributors may be used to endorse or promote products derived from
!    this software without specific prior written permission.
!
! THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
! AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
! IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
! ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
! LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
! CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
! SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
! INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
! CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
! ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
! POSSIBILITY OF SUCH DAMAGE.

program plotqmsd
  use osl_core, wp => osl_wp
  use osl_random
  use dataset
  use msd_model
  implicit none

  !===== User-Defined Settings ===============================================!

  !----- Plotting Options ----------------------------------------------------!

  ! Set OUTPUT to either 'fig', 'eps', or 'pdf' to produce a graphics file
  character(len=3), parameter :: output = 'fig'

  !----- Model Specification -------------------------------------------------!

  ! Finite support of X_1
  real(wp), parameter, dimension(5) :: x1_sup = (/ -2.0_wp, -1.0_wp, 0.0_wp, 1.0_wp, 2.0_wp /)

  ! Finite support of X_2
  real(wp), parameter, dimension(5) :: x2_sup = (/ -2.0_wp, -1.0_wp, 0.0_wp, 1.0_wp, 2.0_wp /)

  ! Population parameters
  real(wp), parameter :: beta0 = -0.5_wp

  !----- Sample Specification ------------------------------------------------!

  integer, parameter :: nn = 50                   ! Number of observations

  !----- Grid Search Options -------------------------------------------------!

  integer, parameter :: n_grid = 4001
  real(wp) :: grid_lb = -2.0_wp
  real(wp) :: grid_ub = 2.0_wp

  type(osl_rng_t) :: rng
  type(msd_model_t) :: msd_mod
  type(dataset_t) :: dat

  !===== Internal Variables ==================================================!

  real(wp), dimension(n_grid) :: beta_grid        ! Grid support
  real(wp), dimension(n_grid) :: Q                ! Population objective fn.
  real(wp), dimension(n_grid) :: Q_n              ! Sample objective fn.
  integer :: k                                    ! Grid point index

  !===== Main Program ========================================================!

  ! Construct a grid for theta
  beta_grid = osl_linspace(grid_lb, grid_ub, n_grid)

  ! Initialize the model
  call msd_model_init(msd_mod, x1_sup, x2_sup, beta0)

  ! Generate data
  call osl_rng_seed(rng, 475827344)
  call dataset_init(dat, nn)
  call msd_model_generate_data(msd_mod, rng, dat)

  ! Evaluate the population and sample objective functions
  do k = 1, n_grid
     call msd_model_population_objective(msd_mod, beta_grid(k), Q(k))
     call msd_model_sample_objective(msd_mod, dat, beta_grid(k), Q_n(k))
  end do

  ! Write data file
  open(UNIT=25, FILE='qmsd.dat')
  do k = 1, n_grid
     write(25, '(3g17.5)') beta_grid(k), Q(k), Q_n(k)
  end do
  close(25)

  ! Generate plots
  select case (output)
  case ('fig')
     print '(a)', 'set terminal fig textspecial'
     print '(a)', 'set output "qmsd.fig"'
  case ('eps')
     print '(a)', 'set terminal postscript eps enhanced monochrome'
     print '(a)', 'set output "qmsd.eps"'
  case ('pdf')
     print '(a)', 'set terminal pdf monochrome'
     print '(a)', 'set output "qmsd.pdf"'
  end select
  print '(a)', 'set size 0.8,0.8'
  print '(a)', 'set xlabel "$\\beta$"'
  print '(a)', 'set key spacing 1.75'
  print '(a)', 'plot' &
       // ' "qmsd.dat" using 1:2 with lines lt 1 title "$Q(\\beta)$",' &
       // ' "qmsd.dat" using 1:3 with steps lt 3 title "$Q_n(\\beta)$"'

  ! Clean up
  call dataset_free(dat)
  call msd_model_free(msd_mod)
end program plotqmsd
